from typing import List, Optional
import functools
import h5py
import matplotlib as mpl
import numpy as np

from fluiddyn.util import mpi
from fluidsim.base.output.spectra import Spectra
from .normal_mode import NormalModeBase


class SpectraSW1L(Spectra):
    """Save and plot spectra."""

    def __init__(self, output):
        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f

        super().__init__(output)

    def _init_online_plot(self):
        super()._init_online_plot()
        if mpi.rank == 0:
            self.axe.set_title("spectra\n" + self.output.summary_simul)

    def compute(self):
        """compute the values at one time."""
        # compute 'quantities_fft'
        (
            energyK_fft,
            energyA_fft,
            energyKr_fft,
        ) = self.output.compute_energies_fft()
        ErtelPE_fft, CharneyPE_fft = self.output.compute_PE_fft()

        # compute the spectra 1D
        spectrum1Dkx_EK, spectrum1Dky_EK = self.spectra1D_from_fft(energyK_fft)
        spectrum1Dkx_EA, spectrum1Dky_EA = self.spectra1D_from_fft(energyA_fft)
        spectrum1Dkx_EKr, spectrum1Dky_EKr = self.spectra1D_from_fft(energyKr_fft)
        spectrum1Dkx_EPE, spectrum1Dky_EPE = self.spectra1D_from_fft(ErtelPE_fft)
        spectrum1Dkx_CPE, spectrum1Dky_CPE = self.spectra1D_from_fft(
            CharneyPE_fft
        )

        dict_spectra1D = {
            "spectrum1Dkx_EK": spectrum1Dkx_EK,
            "spectrum1Dky_EK": spectrum1Dky_EK,
            "spectrum1Dkx_EA": spectrum1Dkx_EA,
            "spectrum1Dky_EA": spectrum1Dky_EA,
            "spectrum1Dkx_EKr": spectrum1Dkx_EKr,
            "spectrum1Dky_EKr": spectrum1Dky_EKr,
            "spectrum1Dkx_EPE": spectrum1Dkx_EPE,
            "spectrum1Dky_EPE": spectrum1Dky_EPE,
            "spectrum1Dkx_CPE": spectrum1Dkx_CPE,
            "spectrum1Dky_CPE": spectrum1Dky_CPE,
        }

        # compute the spectra 2D
        spectrum2D_EK = self.spectrum2D_from_fft(energyK_fft)
        spectrum2D_EA = self.spectrum2D_from_fft(energyA_fft)
        spectrum2D_EKr = self.spectrum2D_from_fft(energyKr_fft)
        spectrum2D_EPE = self.spectrum2D_from_fft(ErtelPE_fft)
        spectrum2D_CPE = self.spectrum2D_from_fft(CharneyPE_fft)

        dict_spectra2D = {
            "spectrum2D_EK": spectrum2D_EK,
            "spectrum2D_EA": spectrum2D_EA,
            "spectrum2D_EKr": spectrum2D_EKr,
            "spectrum2D_EPE": spectrum2D_EPE,
            "spectrum2D_CPE": spectrum2D_CPE,
        }

        dict_lin_spectra1D, dict_lin_spectra2D = self.compute_lin_spectra()
        dict_spectra1D.update(dict_lin_spectra1D)
        dict_spectra2D.update(dict_lin_spectra2D)

        return dict_spectra1D, dict_spectra2D

    def compute_lin_spectra(self):
        (
            energy_glin_fft,
            energy_dlin_fft,
            energy_alin_fft,
        ) = self.output.compute_lin_energies_fft()

        spectrum1Dkx_Eglin, spectrum1Dky_Eglin = self.spectra1D_from_fft(
            energy_glin_fft
        )
        spectrum1Dkx_Edlin, spectrum1Dky_Edlin = self.spectra1D_from_fft(
            energy_dlin_fft
        )
        spectrum1Dkx_Ealin, spectrum1Dky_Ealin = self.spectra1D_from_fft(
            energy_alin_fft
        )

        dict_spectra1D = {
            "spectrum1Dkx_Eglin": spectrum1Dkx_Eglin,
            "spectrum1Dky_Eglin": spectrum1Dky_Eglin,
            "spectrum1Dkx_Edlin": spectrum1Dkx_Edlin,
            "spectrum1Dky_Edlin": spectrum1Dky_Edlin,
            "spectrum1Dkx_Ealin": spectrum1Dkx_Ealin,
            "spectrum1Dky_Ealin": spectrum1Dky_Ealin,
        }

        spectrum2D_Eglin = self.spectrum2D_from_fft(energy_glin_fft)
        spectrum2D_Edlin = self.spectrum2D_from_fft(energy_dlin_fft)
        spectrum2D_Ealin = self.spectrum2D_from_fft(energy_alin_fft)

        dict_spectra2D = {
            "spectrum2D_Eglin": spectrum2D_Eglin,
            "spectrum2D_Edlin": spectrum2D_Edlin,
            "spectrum2D_Ealin": spectrum2D_Ealin,
        }

        return dict_spectra1D, dict_spectra2D

    def _online_plot_saving(self, dict_spectra1D, dict_spectra2D):
        if (
            self.params.oper.nx == self.params.oper.ny
            and self.params.oper.Lx == self.params.oper.Ly
        ):
            spectrum2D_EK = dict_spectra2D["spectrum2D_EK"]
            spectrum2D_EA = dict_spectra2D["spectrum2D_EA"]
            spectrum2D_EKr = dict_spectra2D["spectrum2D_EKr"]
            spectrum2D_E = spectrum2D_EK + spectrum2D_EA
            spectrum2D_EKd = spectrum2D_EK - spectrum2D_EKr
            khE = self.oper.khE
            coef_norm = khE ** (3.0)
            self.axe.loglog(khE, spectrum2D_E * coef_norm, "k")
            self.axe.loglog(khE, spectrum2D_EK * coef_norm, "r")
            self.axe.loglog(khE, spectrum2D_EA * coef_norm, "b")
            self.axe.loglog(khE, spectrum2D_EKr * coef_norm, "r--")
            self.axe.loglog(khE, spectrum2D_EKd * coef_norm, "r:")
            lin_inf, lin_sup = self.axe.get_ylim()
            if lin_inf < 10e-6:
                lin_inf = 10e-6
            self.axe.set_ylim([lin_inf, lin_sup])
        else:
            print(
                "you need to implement the ploting "
                "of the spectra for this case"
            )

    def plot1d(
        self,
        tmin: float = 0,
        tmax: float = 1000,
        delta_t: float = 2,
        coef_compensate: float = 3,
        coef_norm: Optional[np.ndarray] = None,
        ax: Optional[mpl.axes.Axes] = None,
        help_lines: bool = True,
    ):

        with h5py.File(self.path_file1D, "r") as h5file:
            dset_times = h5file["times"]
            times = dset_times[...]
            # nb_spectra = times.shape[0]

            dset_kxE = h5file["kxE"]
            # dset_kyE = h5file['kyE']

            kh = dset_kxE[...]

            dset_spectrum1Dkx_EK = h5file["spectrum1Dkx_EK"]
            dset_spectrum1Dky_EK = h5file["spectrum1Dky_EK"]
            dset_spectrum1Dkx_EA = h5file["spectrum1Dkx_EA"]
            dset_spectrum1Dky_EA = h5file["spectrum1Dky_EA"]

            # dset_spectrum1Dkx_EKr = h5file["spectrum1Dkx_EKr"]
            # dset_spectrum1Dky_EKr = h5file["spectrum1Dky_EKr"]

            # nt = len(times)

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot1d(tmin={}, tmax={}, delta_t={:.2f},".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = """plot 1D spectra
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            if ax is None:
                fig, ax = self.output.figure_axe()

            ax.set_xlabel("$k_h$")
            ax.set_ylabel("1D spectra")
            ax.set_title("1D spectra\n" + self.output.summary_simul)
            ax.set_xscale("log")
            ax.set_yscale("log")

            if coef_norm is None:
                coef_norm = kh ** (coef_compensate)

            # min_to_plot = 1e-16

            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    E_K = dset_spectrum1Dkx_EK[it] + dset_spectrum1Dky_EK[it]
                    # E_K[E_K<min_to_plot] = 0.
                    E_A = dset_spectrum1Dkx_EA[it] + dset_spectrum1Dky_EA[it]
                    # E_A[E_A<min_to_plot] = 0.
                    E_tot = E_K + E_A

                    # E_Kr = dset_spectrum1Dkx_EKr[it] + dset_spectrum1Dky_EKr[it]
                    # E_Kr[E_Kr<min_to_plot] = 0.
                    # E_Kd = E_K - E_Kr

                    ax.plot(kh, E_tot * coef_norm, "k", linewidth=2)
                    ax.plot(kh, E_K * coef_norm, "r", linewidth=1)
                    ax.plot(kh, E_A * coef_norm, "b", linewidth=1)
            # ax.plot(kh, E_Kr*coef_norm, 'r--', linewidth=1)
            # ax.plot(kh, E_Kd*coef_norm, 'r:', linewidth=1)

            E_K = (
                dset_spectrum1Dkx_EK[imin_plot : imax_plot + 1]
                + dset_spectrum1Dky_EK[imin_plot : imax_plot + 1]
            ).mean(0)

            E_A = (
                dset_spectrum1Dkx_EA[imin_plot : imax_plot + 1]
                + dset_spectrum1Dky_EA[imin_plot : imax_plot + 1]
            ).mean(0)

        ax.plot(kh, E_K * coef_norm, "r", linewidth=2)
        ax.plot(kh, E_A * coef_norm, "b", linewidth=2)

        if help_lines:
            kh_pos = kh[kh > 0]
            coef_norm = coef_norm[kh > 0]
            ax.plot(kh_pos, kh_pos ** (-3) * coef_norm, "k--", linewidth=1)
            ax.plot(kh_pos, kh_pos ** (-5.0 / 3) * coef_norm, "k-.", linewidth=1)

    def plot2d(
        self,
        tmin: float = 0,
        tmax: float = 1000,
        delta_t: float = 2,
        coef_compensate: float = 3,
        coef_norm: Optional[np.ndarray] = None,
        keys: List[str] = ["Etot", "EK", "EA", "EKr", "EKd"],
        colors: List[str] = ["k", "r", "b", "r--", "r:"],
        kh_norm: float = 1,
        ax: Optional[mpl.axes.Axes] = None,
        help_lines: bool = True,
    ):

        with h5py.File(self.path_file2D, "r") as h5file:
            times = h5file["times"][...]

            dset_khE = h5file["khE"]
            kh = dset_khE[...] / kh_norm

            dset_spectrumEK = h5file["spectrum2D_EK"]
            dset_spectrumEA = h5file["spectrum2D_EA"]
            dset_spectrumEKr = h5file["spectrum2D_EKr"]

            delta_t_save = np.mean(times[1:] - times[0:-1])
            delta_i_plot = int(np.round(delta_t / delta_t_save))
            if delta_i_plot == 0 and delta_t != 0.0:
                delta_i_plot = 1
            delta_t = delta_i_plot * delta_t_save

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = "plot2d(tmin={}, tmax={}, delta_t={:.2f},".format(
                tmin, tmax, delta_t
            )
            print(to_print)

            to_print = """plot 2D spectra
    tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
    imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
                tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
            )
            print(to_print)

            if ax is None:
                fig, ax = self.output.figure_axe()

            ax.set_xlabel("$k_h$")
            ax.set_ylabel("2D spectra")
            ax.set_title("2D spectra\n" + self.output.summary_simul)
            ax.set_xscale("log")
            ax.set_yscale("log")

            if coef_norm is None:
                coef_norm = kh ** coef_compensate

            machine_zero = 1e-15
            if delta_t != 0.0:
                for it in range(imin_plot, imax_plot + 1, delta_i_plot):
                    for k, c in zip(keys, colors):
                        dset = self._get_field_to_plot(it, k, h5file)
                        dset[dset < 10e-16] = machine_zero
                        ax.plot(kh, dset * coef_norm, c, linewidth=1)

            EK = dset_spectrumEK[imin_plot : imax_plot + 1].mean(0)
            EA = dset_spectrumEA[imin_plot : imax_plot + 1].mean(0)
            EKr = dset_spectrumEKr[imin_plot : imax_plot + 1].mean(0)

            EK[abs(EK) < 10e-16] = machine_zero
            EA[abs(EA) < 10e-16] = machine_zero
            EKr[abs(EKr) < 10e-16] = machine_zero

            E_tot = EK + EA
            EKd = EK - EKr + machine_zero

            if "Etot" in keys:
                ax.plot(
                    kh,
                    E_tot * coef_norm,
                    colors[0],
                    linewidth=2,
                    label="$E_{tot}$",
                )

            if "EK" in keys:
                ax.plot(kh, EK * coef_norm, "r", linewidth=1, label="$E_{K}$")
                ax.plot(kh, -EK * coef_norm, "k-", linewidth=1)

            if "EA" in keys:
                ax.plot(kh, EA * coef_norm, "b", linewidth=1, label="$E_{A}$")

            if "EKr" in keys:
                ax.plot(kh, EKr * coef_norm, "r--", linewidth=1, label="$E_{Kr}$")

            if "EKd" in keys:
                ax.plot(kh, EKd * coef_norm, "r:", linewidth=1, label="$E_{Kd}$")
                ax.plot(kh, -EKd * coef_norm, "k:", linewidth=1)

            self._plot2d_lin_spectra(
                h5file, ax, imin_plot, imax_plot, kh, coef_norm, keys
            )

        if help_lines:
            kh_pos = kh[kh > 0]
            coef_norm = coef_norm[kh > 0]
            ax.plot(kh_pos, kh_pos ** (-2) * coef_norm, "k-", linewidth=1)
            ax.plot(kh_pos, kh_pos ** (-3) * coef_norm, "k--", linewidth=1)
            ax.plot(kh_pos, kh_pos ** (-5.0 / 3) * coef_norm, "k-.", linewidth=1)

            postxt = kh.max()
            ax.text(postxt, postxt ** (-2 + coef_compensate), r"$k^{-2}$")
            ax.text(postxt, postxt ** (-3 + coef_compensate), r"$k^{-3}$")
            ax.text(postxt, postxt ** (-5.0 / 3 + coef_compensate), r"$k^{-5/3}$")

        ax.legend()

    def plot_diss(
        self,
        tmin=0,
        tmax=1000,
        delta_t=2,
        keys=["Dtot", "DK", "DA", "DKr", "DKd"],
        colors=["k", "r", "b", "r--", "r:"],
        kh_norm=1,
        ax=None,
    ):
        """Plot the dissipation spectra."""

        def get_nu(o):
            return getattr(self.sim.params, f"nu_{o}")

        for order in [2, 4, 8]:
            nu = get_nu(order)
            if nu > 0:
                break
        else:
            raise ValueError("Viscosity is zero?")

        with h5py.File(self.path_file2D, "r") as h5file:
            dset_khE = h5file["khE"]
            kh = dset_khE[...]

        coef_norm = 2 * nu * kh ** order
        keys = ["E" + k.lstrip("D") for k in keys]
        self.plot2d(
            tmin, tmax, delta_t, 0, coef_norm, keys, colors, kh_norm, ax, False
        )

    def _plot2d_lin_spectra(
        self, h5file, ax, imin_plot, imax_plot, kh, coef_norm, keys
    ):
        machine_zero = 1e-15
        if self.sim.info.solver.short_name.startswith("SW1L"):
            dset_spectrumEdlin = h5file["spectrum2D_Edlin"]
            Edlin = (
                dset_spectrumEdlin[imin_plot : imax_plot + 1].mean(0)
                + machine_zero
            )
            ax.plot(kh, Edlin * coef_norm, "c", linewidth=1, label="$E_{D}$")

        if self.params.f != 0:
            dset_spectrumEglin = h5file["spectrum2D_Eglin"]
            Eglin = (
                dset_spectrumEglin[imin_plot : imax_plot + 1].mean(0)
                + machine_zero
            )
            ax.plot(kh, Eglin * coef_norm, "g", linewidth=1, label="$E_{G}$")

            dset_spectrumEalin = h5file["spectrum2D_Ealin"]
            Ealin = (
                dset_spectrumEalin[imin_plot : imax_plot + 1].mean(0)
                + machine_zero
            )
            ax.plot(kh, Ealin * coef_norm, "y", linewidth=1, label="$E_{A}$")

    def _get_field_to_plot(self, idx, key_field=None, h5file=None):
        if key_field is None:
            key_field = self._ani_key

        def select(idx, key_field, h5file):
            if key_field == "Etot" or key_field is None:
                self._ani_key = "Etot"
                y = h5file["spectrum2D_EK"][idx] + h5file["spectrum2D_EA"][idx]
            elif key_field == "EKd":
                y = h5file["spectrum2D_EK"][idx] - h5file["spectrum2D_EKr"][idx]
            else:
                try:
                    key_field = "spectrum2D_" + key_field
                    y = h5file[key_field][idx]
                except:
                    raise KeyError("Unknown key ", key_field)

            return y

        if h5file is None:
            with h5py.File(self.path_file2D) as h5file:
                return select(idx, key_field, h5file)

        else:
            return select(idx, key_field, h5file)


class SpectraSW1LNormalMode(SpectraSW1L):
    def __init__(self, output):
        self.norm_mode = NormalModeBase(output)
        super().__init__(output)

    def compute_lin_spectra(self):
        (
            energy_glin_fft,
            energy_aplin_fft,
            energy_amlin_fft,
        ) = self.norm_mode.compute_qapam_energies_fft()

        energy_alin_fft = energy_aplin_fft + energy_amlin_fft
        spectrum1Dkx_Eglin, spectrum1Dky_Eglin = self.spectra1D_from_fft(
            energy_glin_fft
        )
        spectrum1Dkx_Ealin, spectrum1Dky_Ealin = self.spectra1D_from_fft(
            energy_alin_fft
        )

        dict_spectra1D = {
            "spectrum1Dkx_Eglin": spectrum1Dkx_Eglin,
            "spectrum1Dky_Eglin": spectrum1Dky_Eglin,
            "spectrum1Dkx_Ealin": spectrum1Dkx_Ealin,
            "spectrum1Dky_Ealin": spectrum1Dky_Ealin,
        }

        spectrum2D_Eglin = self.spectrum2D_from_fft(energy_glin_fft)
        spectrum2D_Ealin = self.spectrum2D_from_fft(energy_alin_fft)

        dict_spectra2D = {
            "spectrum2D_Eglin": spectrum2D_Eglin,
            "spectrum2D_Ealin": spectrum2D_Ealin,
        }

        return dict_spectra1D, dict_spectra2D

    def plot2d(
        self,
        tmin=0,
        tmax=1000,
        delta_t=2,
        coef_compensate=3,
        coef_norm=None,
        keys=["Etot", "EK", "Eglin", "Ealin"],
        colors=["k", "r", "g", "y"],
        kh_norm=1,
        ax=None,
        help_lines=True,
    ):
        # Ideally functool.partialmethod would suffice, but issues due to mixing args and kwargs
        super().plot2d(
            tmin,
            tmax,
            delta_t,
            coef_compensate,
            coef_norm,
            keys,
            colors,
            kh_norm,
            ax,
            help_lines,
        )

    def _plot2d_lin_spectra(
        self, h5file, ax, imin_plot, imax_plot, kh, coef_norm, keys
    ):
        machine_zero = 1e-15
        if self.sim.info.solver.short_name.startswith("SW1L"):
            if "spectrum2D_Edlin" in h5file.keys():
                dset_spectrumEalin = h5file[
                    "spectrum2D_Edlin"
                ]  # TODO: To be removed. Kept for compatibility
            else:
                dset_spectrumEalin = h5file["spectrum2D_Ealin"]

            if "Ealin" in keys:
                Ealin = (
                    dset_spectrumEalin[imin_plot : imax_plot + 1].mean(0)
                    + machine_zero
                )
                ax.plot(
                    kh, Ealin * coef_norm, "y", linewidth=1, label="$E_{AGEO}$"
                )

            if "Eglin" in keys:
                dset_spectrumEglin = h5file["spectrum2D_Eglin"]
                Eglin = (
                    dset_spectrumEglin[imin_plot : imax_plot + 1].mean(0)
                    + machine_zero
                )
                ax.plot(
                    kh, Eglin * coef_norm, "g", linewidth=1, label="$E_{GEO}$"
                )



"""Time stepping (:mod:`fluidsim.base.basilisk.time_stepping`)
==============================================================


Provides:

.. autoclass:: TimeSteppingBasilisk
   :members:
   :private-members:

"""

from time import time
from signal import signal

import numpy as np


class TimeSteppingBasilisk:
    """Time stepping class to handle Basilisk's event loop and FluidSim output."""

    @staticmethod
    def _complete_params_with_default(params):
        """This static method is used to complete the *params* container."""
        attribs = {"USE_T_END": True, "t_end": 10.0, "it_end": 10, "deltat0": 0.5}
        params._set_child("time_stepping", attribs=attribs)

    def __init__(self, sim):
        self.params = sim.params
        self.sim = sim

        self.it = 0
        self.t = 0

        self._has_to_stop = False

        def handler_signals(signal_number, stack):
            print(f"signal {signal_number} received.")
            self._has_to_stop = True

        signal(12, handler_signals)

    def start(self):
        output = self.sim.output
        output.init_with_initialized_state()

        print_stdout = output.print_stdout
        print_stdout(
            "*************************************\n"
            + "Beginning of the computation"
        )
        if output._has_to_save:
            output.phys_fields.save()

        params_ts = self.sim.params.time_stepping
        if params_ts.USE_T_END:
            t_end = params_ts.t_end
            nb_time_steps = int(float(params_ts.t_end) / params_ts.deltat0)
            self.deltat = t_end / nb_time_steps
        else:
            nb_time_steps = params_ts.it_end
            t_end = params_ts.deltat0 * params_ts.it_end
            self.deltat = params_ts.deltat0

        def one_time_step(i, t):
            print(f"Basilisk one_time_step: (i, t) = {i}, {t}")
            self.t = t
            self.it = i
            self.sim.state._get_state_from_basilisk()
            self.sim.output.one_time_step()

        print("t_end, nb_time_steps", t_end, nb_time_steps)

        self.sim.basilisk.event(
            one_time_step, t=np.linspace(0.0, t_end, nb_time_steps + 1)
        )

        time_begining_simul = time()

        self.sim.basilisk.run()

        total_time_simul = time() - time_begining_simul
        output.end_of_simul(total_time_simul)



""" """

import numpy as np

from fluidsim.solvers.sw1l.output import OutputBaseSW1L


class OutputSW1LExactlinModified(OutputBaseSW1L):
    """subclass :class:`OutputSW1L`"""

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        OutputBaseSW1L._complete_info_solver(info_solver)

        classes = info_solver.classes.Output.classes

        classes.SpatialMeans.class_name = "SpatialMeansMSW1L"
        classes.SpectralEnergyBudget.class_name = (
            "SpectralEnergyBudgetSW1LModified"
        )

    def compute_energies_fft(self):
        ux_fft = self.sim.state.get_var("ux_fft")
        uy_fft = self.sim.state.get_var("uy_fft")
        eta_fft = self.sim.state.get_var("eta_fft")
        energyA_fft = self.sim.params.c2 * np.abs(eta_fft) ** 2 / 2
        energyK_fft = np.abs(ux_fft) ** 2 / 2 + np.abs(uy_fft) ** 2 / 2
        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        uxr_fft, uyr_fft = self.oper.vecfft_from_rotfft(rot_fft)
        energyKr_fft = np.abs(uxr_fft) ** 2 / 2 + np.abs(uyr_fft) ** 2 / 2
        return energyK_fft, energyA_fft, energyKr_fft

    def compute_energiesKA_fft(self):
        ux_fft = self.sim.state.get_var("ux_fft")
        uy_fft = self.sim.state.get_var("uy_fft")
        eta_fft = self.sim.state.get_var("eta_fft")
        energyA_fft = self.sim.params.c2 * np.abs(eta_fft) ** 2 / 2
        energyK_fft = np.abs(ux_fft) ** 2 / 2 + np.abs(uy_fft) ** 2 / 2
        return energyK_fft, energyA_fft

    def compute_PV_fft(self):
        # compute Ertel and Charney (QG) potential vorticity
        rot = self.sim.state.get_var("rot")
        eta = self.sim.state.state_phys.get_var("eta")
        ErtelPV_fft = self.oper.fft2((self.sim.params.f + rot) / (1.0 + eta))
        ux_fft = self.sim.state.get_var("ux_fft")
        uy_fft = self.sim.state.get_var("uy_fft")
        rot_fft = self.oper.rotfft_from_vecfft(ux_fft, uy_fft)
        eta_fft = self.sim.state.get_var("eta_fft")
        CharneyPV_fft = rot_fft - self.sim.params.f * eta_fft
        return ErtelPV_fft, CharneyPV_fft



"""Spectra (:mod:`fluidsim.solvers.ns3d.strat.output.spectra`)
==============================================================

.. autoclass:: SpectraNS3DStrat
   :members:
   :private-members:

"""

import numpy as np
import h5py

from fluidsim.solvers.ns3d.output.spectra import (
    SpectraNS3D,
    _get_averaged_spectrum,
)


class SpectraNS3DStrat(SpectraNS3D):
    """Save and plot spectra."""

    def compute(self):
        """compute the values at one time."""

        get_var = self.sim.state.state_spect.get_var
        b_fft = get_var("b_fft")
        vx_fft = get_var("vx_fft")
        vy_fft = get_var("vy_fft")
        vz_fft = get_var("vz_fft")

        urx_fft, ury_fft, udx_fft, udy_fft = self.sim.oper.urudfft_from_vxvyfft(
            vx_fft, vy_fft
        )

        nrj_vx_fft = 0.5 * np.abs(vx_fft) ** 2
        nrj_vy_fft = 0.5 * np.abs(vy_fft) ** 2
        nrj_vz_fft = 0.5 * np.abs(vz_fft) ** 2

        nrj_A_fft = 0.5 / self.sim.params.N ** 2 * np.abs(b_fft) ** 2
        nrj_Khr_fft = 0.5 * (np.abs(urx_fft) ** 2 + np.abs(ury_fft) ** 2)
        nrj_Khd_fft = 0.5 * (np.abs(udx_fft) ** 2 + np.abs(udy_fft) ** 2)

        s_vx_kx, s_vx_ky, s_vx_kz = self.oper.compute_1dspectra(nrj_vx_fft)
        s_vy_kx, s_vy_ky, s_vy_kz = self.oper.compute_1dspectra(nrj_vy_fft)
        s_vz_kx, s_vz_ky, s_vz_kz = self.oper.compute_1dspectra(nrj_vz_fft)

        s_A_kx, s_A_ky, s_A_kz = self.oper.compute_1dspectra(nrj_A_fft)
        s_Khr_kx, s_Khr_ky, s_Khr_kz = self.oper.compute_1dspectra(nrj_Khr_fft)
        s_Khd_kx, s_Khd_ky, s_Khd_kz = self.oper.compute_1dspectra(nrj_Khd_fft)

        s_kx = s_vx_kx + s_vy_kx + s_vz_kx
        s_ky = s_vx_ky + s_vy_ky + s_vz_ky
        s_kz = s_vx_kz + s_vy_kz + s_vz_kz

        dict_spectra1d = {
            "vx_kx": s_vx_kx,
            "vx_ky": s_vx_ky,
            "vx_kz": s_vx_kz,
            "vy_kx": s_vy_kx,
            "vy_ky": s_vy_ky,
            "vy_kz": s_vy_kz,
            "vz_kx": s_vz_kx,
            "vz_ky": s_vz_ky,
            "vz_kz": s_vz_kz,
            "E_kx": s_kx,
            "E_ky": s_ky,
            "E_kz": s_kz,
            "A_kx": s_A_kx,
            "A_ky": s_A_ky,
            "A_kz": s_A_kz,
            "Khr_kx": s_Khr_kx,
            "Khr_ky": s_Khr_ky,
            "Khr_kz": s_Khr_kz,
            "Khd_kx": s_Khd_kx,
            "Khd_ky": s_Khd_ky,
            "Khd_kz": s_Khd_kz,
        }
        dict_spectra1d = {"spectra_" + k: v for k, v in dict_spectra1d.items()}

        s_vx = self.oper.compute_3dspectrum(nrj_vx_fft)
        s_vy = self.oper.compute_3dspectrum(nrj_vy_fft)
        s_vz = self.oper.compute_3dspectrum(nrj_vz_fft)
        s_A = self.oper.compute_3dspectrum(nrj_A_fft)
        s_Khr = self.oper.compute_3dspectrum(nrj_Khr_fft)
        s_Khd = self.oper.compute_3dspectrum(nrj_Khd_fft)
        dict_spectra3d = {
            "vx": s_vx,
            "vy": s_vy,
            "vz": s_vz,
            "E": s_vx + s_vy + s_vy,
            "A": s_A,
            "Khr": s_Khr,
            "Khd": s_Khd,
        }
        dict_spectra3d = {"spectra_" + k: v for k, v in dict_spectra3d.items()}

        if self.has_to_save_kzkh():
            dict_kzkh = {
                "A": self.oper.compute_spectrum_kzkh(nrj_A_fft),
                "Khr": self.oper.compute_spectrum_kzkh(nrj_Khr_fft),
                "Khd": self.oper.compute_spectrum_kzkh(nrj_Khd_fft),
                "Kz": self.oper.compute_spectrum_kzkh(nrj_vz_fft),
            }
        else:
            dict_kzkh = None

        return dict_spectra1d, dict_spectra3d, dict_kzkh

    def plot_kzkh(self, tmin=0, tmax=None, key="Khd", ax=None):
        super().plot_kzkh(tmin, tmax, key, ax)

    def _plot1d_direction(
        self, direction, imin_plot, imax_plot, coef_compensate, ax
    ):

        with h5py.File(self.path_file1d, "r") as h5file:
            ks = h5file["k" + direction][...]

            def _get_spectrum(key):
                return _get_averaged_spectrum(
                    key + direction, h5file, imin_plot, imax_plot
                )

            spectrumK = _get_spectrum("spectra_E_k")
            spectrumA = _get_spectrum("spectra_A_k")
            spectrumKhd = _get_spectrum("spectra_Khd_k")
            spectrumKz = _get_spectrum("spectra_vz_k")

        ks_no0 = ks.copy()
        ks_no0[ks == 0] = np.nan
        coef_norm = ks_no0 ** (coef_compensate)

        style_line = ""
        if direction == "z":
            style_line = ":"

        def _plot(spectrum, color, label, linewidth=2):
            ax.plot(
                ks,
                spectrum * coef_norm,
                color + style_line,
                linewidth=linewidth,
                label=label,
            )

        _plot(spectrumK, "r", f"$E_K(k_{direction})$")
        _plot(spectrumA, "b", f"$E_A(k_{direction})$")
        _plot(spectrumKhd + spectrumKz, "m", "poloidal", 1)

    def plot_kzkh_cumul_diss(self, tmin=0, tmax=None):
        path_file = self.path_file_kzkh
        with h5py.File(path_file, "r") as h5file:
            times = h5file["times"][...]
        if tmax is None:
            tmax = times.max()

        # load kzkh spectra
        spectra_K = 0
        data = self.load_kzkh_mean(tmin=tmin, tmax=tmax, key_to_load="Khd")
        spectra_K += data["Khd"]
        data = self.load_kzkh_mean(tmin=tmin, tmax=tmax, key_to_load="Khr")
        spectra_K += data["Khr"]
        data = self.load_kzkh_mean(tmin=tmin, tmax=tmax, key_to_load="Kz")
        spectra_K += data["Kz"]

        # spectral space
        kh = data["kh_spectra"]
        kz = data["kz"]
        deltakh = kh[1]
        deltakz = kz[1]
        KH, KZ = np.meshgrid(kh, kz)
        K2 = KH ** 2 + KZ ** 2
        K4 = K2 ** 2

        del KH, KZ

        # nu_2 dissipation fluxes
        fd = self.sim.params.nu_2 * K2
        diss = fd * spectra_K
        hflux_diss_nu2 = deltakz * diss.sum(0)
        hflux_diss_nu2 = deltakh * np.cumsum(hflux_diss_nu2)
        zflux_diss_nu2 = deltakh * diss.sum(1)
        zflux_diss_nu2 = deltakz * np.cumsum(zflux_diss_nu2)

        # nu_4 dissipation fluxes
        fd = self.sim.params.nu_4 * K4
        diss = fd * spectra_K
        hflux_diss_nu4 = deltakz * diss.sum(0)
        hflux_diss_nu4 = deltakh * np.cumsum(hflux_diss_nu4)
        zflux_diss_nu4 = deltakh * diss.sum(1)
        zflux_diss_nu4 = deltakz * np.cumsum(zflux_diss_nu4)

        del fd, diss, K2, K4

        # normalize by total dissipation
        eps_tot = hflux_diss_nu2[-1] + hflux_diss_nu4[-1]
        hflux_diss_nu2 /= eps_tot
        zflux_diss_nu2 /= eps_tot
        hflux_diss_nu4 /= eps_tot
        zflux_diss_nu4 /= eps_tot

        # plot
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(r"$k$")
        ax.set_ylabel(r"$D(k)/\epsilon$")

        ax.plot(kh, hflux_diss_nu2, "r-", label=r"$D_2(k_h)$")
        ax.plot(kz, zflux_diss_nu2, "r:", label=r"$D_2(k_z)$")
        ax.plot(kh, hflux_diss_nu4, "m-", label=r"$D_4(k_h)$")
        ax.plot(kz, zflux_diss_nu4, "m:", label=r"$D_4(k_z)$")

        ax.set_title(
            f"kzkh cumulative dissipation spectra (tmin={tmin:.2g}, tmax={tmax:.2g})\n"
            + self.output.summary_simul
        )
        ax.set_xscale("log")
        fig.legend()



"""Plate2d state (:mod:`fluidsim.solvers.plate2d.state`)
==============================================================
"""

import numpy as np

from fluidsim.base.state import StatePseudoSpectral

from fluiddyn.util import mpi


class StatePlate2D(StatePseudoSpectral):
    """Contains the variables corresponding to the state and handles the
    access to other fields for the solver PLATE2D.

    """

    @staticmethod
    def _complete_info_solver(info_solver):
        """Complete the ParamContainer info_solver.

        This is a static method!
        """
        info_solver.classes.State._set_attribs(
            {
                "keys_state_spect": ["w_fft", "z_fft"],
                "keys_state_phys": ["w", "z"],
                "keys_computable": ["chi_fft", "chi", "Nw_fft", "lapz_fft"],
                "keys_phys_needed": ["w", "z"],
                "keys_linear_eigenmodes": [],
            }
        )

    def compute(self, key, SAVE_IN_DICT=True, RAISE_ERROR=True):
        oper = self.oper
        it = self.sim.time_stepping.it
        if key in self.vars_computed and it == self.it_computed[key]:
            return self.vars_computed[key]

        if key == "chi_fft":
            z_fft = self.state_spect.get_var("z_fft")
            mamp_zz = oper.monge_ampere_from_fft(z_fft, z_fft)
            result = oper.invlaplacian_fft(
                oper.fft2(mamp_zz), order=4, negative=True
            )
        elif key == "chi":
            chi_fft = self.compute("chi_fft")
            result = oper.ifft2(chi_fft)
        elif key == "Nw_fft":
            mamp_zchi = oper.monge_ampere_from_fft(
                self.state_spect.get_var("z_fft"), self.compute("chi_fft")
            )
            result = oper.fft2(mamp_zchi)
        elif key == "lapz_fft":
            z_fft = self.state_spect.get_var("z_fft")
            result = oper.laplacian_fft(z_fft, order=4)
        else:
            to_print = 'Do not know how to compute "' + key + '".'
            if RAISE_ERROR:
                raise ValueError(to_print)

            else:
                if mpi.rank == 0:
                    print(to_print + "\nreturn an array of zeros.")

                result = self.oper.create_arrayX(value=0.0)

        if SAVE_IN_DICT:
            self.vars_computed[key] = result
            self.it_computed[key] = it

        return result

    def statephys_from_statespect(self):
        w_fft = self.state_spect.get_var("w_fft")
        z_fft = self.state_spect.get_var("z_fft")

        w = self.state_phys.get_var("w")
        z = self.state_phys.get_var("z")

        self.oper.ifft_as_arg(w_fft, w)
        self.oper.ifft_as_arg(z_fft, z)

    def init_state_from_wz_fft(self, w_fft, z_fft):
        self.oper.dealiasing(w_fft, z_fft)
        self.state_spect.set_var("w_fft", w_fft)
        self.state_spect.set_var("z_fft", z_fft)
        self.statephys_from_statespect()

    def init_statespect_from(self, **kwargs):
        if len(kwargs) == 1:
            if "w_fft" in kwargs:
                w_fft = kwargs["w_fft"]
                z_fft = np.zeros_like(w_fft)
                self.init_state_from_wz_fft(w_fft, z_fft)
        else:
            super().init_statespect_from(**kwargs)



import unittest

import fluiddyn as fld
from fluiddyn.util import mpi

# to get fld.show
import fluiddyn.output

from fluidsim.base.solvers.base import SimulBase as Simul

from fluidsim.util.testing import TestSimul


@unittest.skipIf(mpi.nb_proc > 1, "Od solvers do not work with mpi")
class TestBaseSolver(TestSimul):
    Simul = Simul

    def test_simul(self):
        """Should be able to run a base experiment."""
        self.sim.time_stepping.start()

        fld.show()


if __name__ == "__main__":
    unittest.main()



import h5py
import numpy as np

from fluidsim.base.output.increments import Increments


class IncrementsSW1L(Increments):
    """A :class:`Increments` object handles the saving of pdf of
    increments.
    """

    def __init__(self, output):
        super().__init__(output)
        params = output.sim.params
        self.c2 = params.c2
        self.f = params.f

    def _online_plot_saving(self, dict_results, key="eta"):
        """online plot on pdf"""
        super()._online_plot_saving(dict_results, key=key)

    def compute(self):
        dict_results = super().compute()

        get_var = self.sim.state.get_var
        ux = get_var("ux")
        uy = get_var("uy")
        eta = get_var("eta")
        Jx = (1 + eta) * ux

        S_uL2JL = np.empty([self.nrx])
        S_uT2JL = np.empty([self.nrx])
        S_c2h2uL = np.empty([self.nrx])
        S_uT2uL = np.empty([self.nrx])

        for irx, rx in enumerate(self.rxs):
            inc_ux = self.oper.compute_increments_dim1(ux, rx)
            inc_uy = self.oper.compute_increments_dim1(uy, rx)
            inc_eta = self.oper.compute_increments_dim1(eta, rx)
            inc_Jx = self.oper.compute_increments_dim1(Jx, rx)
            inc_uy2 = inc_uy ** 2
            S_uL2JL[irx] = np.mean(inc_ux ** 2 * inc_Jx)
            S_uT2JL[irx] = np.mean(inc_uy2 * inc_Jx)
            S_c2h2uL[irx] = self.params.c2 * np.mean(inc_eta ** 2 * inc_ux)
            S_uT2uL[irx] = np.mean(inc_uy2 * inc_ux)

        dict_results["struc_func_uL2JL"] = S_uL2JL
        dict_results["struc_func_uT2JL"] = S_uT2JL
        dict_results["struc_func_c2h2uL"] = S_c2h2uL
        dict_results["struc_func_Kolmo"] = S_uL2JL + S_uT2JL + S_c2h2uL

        dict_results["struc_func_uT2uL"] = S_uT2uL

        return dict_results

    def plot(self, tmin=0, tmax=None, delta_t=2, order=2, yscale="log"):
        """Plot some structure functions."""
        with h5py.File(self.path_file, "r") as h5file:
            times = h5file["times"][...]
            # nt = len(times)

            if tmax is None:
                tmax = times.max()

            rxs = h5file["rxs"][...]

            oper = h5file["/info_simul/params/oper"]
            nx = oper.attrs["nx"]
            Lx = oper.attrs["Lx"]
            deltax = Lx / nx

        rxs = np.array(rxs, dtype=np.float64) * deltax

        # orders = h5file['orders'][...]
        # dset_struc_func_ux = h5file['struc_func_ux']
        # dset_struc_func_uy = h5file['struc_func_uy']

        if len(times) > 1:
            delta_t_save = np.mean(times[1:] - times[0:-1])
        else:
            delta_t_save = delta_t

        delta_i_plot = int(np.round(delta_t / delta_t_save))
        if delta_i_plot == 0 and delta_t != 0.0:
            delta_i_plot = 1
        delta_t = delta_i_plot * delta_t_save

        imin_plot = np.argmin(abs(times - tmin))
        imax_plot = np.argmin(abs(times - tmax))

        tmin_plot = times[imin_plot]
        tmax_plot = times[imax_plot]

        to_print = "plot(tmin={}, tmax={}, delta_t={:.2f})".format(
            tmin, tmax, delta_t
        )
        print(to_print)

        to_print = """plot structure functions
tmin = {:8.6g} ; tmax = {:8.6g} ; delta_t = {:8.6g}
imin = {:8d} ; imax = {:8d} ; delta_i = {:8d}""".format(
            tmin_plot, tmax_plot, delta_t, imin_plot, imax_plot, delta_i_plot
        )
        print(to_print)

        pdf_eta, values_inc_eta, nb_rx_to_plot = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var="eta"
        )

        pdf_ux, values_inc_ux, nb_rx_to_plot = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var="ux"
        )

        pdf_uy, values_inc_uy, nb_rx_to_plot = self.load_pdf_from_file(
            tmin=tmin, tmax=tmax, key_var="uy"
        )

        # iorder = self.iorder_from_order(order)
        order = float(order)

        x_left_axe = 0.12
        z_bottom_axe = 0.56
        width_axe = 0.85
        height_axe = 0.37
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("$r_x$")
        ax1.set_ylabel(r"$\langle \delta u^{" + f"{order}" + "} \\rangle$")
        ax1.set_title("struct. functions\n" + self.output.summary_simul)

        ax1.set_xscale("log")
        ax1.set_yscale(yscale)

        # So_eta = self.strfunc_from_pdf(pdf_eta, values_inc_eta, order)
        So_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, order)
        So_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, order)

        norm = rxs

        # ax1.set_ylabel('struct. functions, order = {0}'.format(order))
        # if delta_t != 0.:
        #     for it in range(imin_plot,imax_plot+1,delta_i_plot):
        #         struc_func_ux = dset_struc_func_ux[it]
        #         struc_func_ux = struc_func_ux.reshape(
        #             [self.norders, self.nrx])
        #         struc_func_uy = dset_struc_func_uy[it]
        #         struc_func_uy = struc_func_uy.reshape(
        #             [self.norders, self.nrx])

        #         ax1.plot(rxs, struc_func_ux[iorder], 'c', linewidth=1)
        #         ax1.plot(rxs, struc_func_uy[iorder], 'm', linewidth=1)

        # struc_func_ux = dset_struc_func_ux[imin_plot:imax_plot+1].mean(0)
        # struc_func_ux = struc_func_ux.reshape([self.norders, self.nrx])
        # struc_func_uy = dset_struc_func_uy[imin_plot:imax_plot+1].mean(0)
        # struc_func_uy = struc_func_uy.reshape([self.norders, self.nrx])

        # ax1.plot(rxs, struc_func_ux[iorder]/norm, 'c', linewidth=2)
        # ax1.plot(rxs, struc_func_uy[iorder]/norm, 'm', linewidth=2)

        ax1.plot(rxs, So_ux / norm, "c-.", linewidth=2)
        ax1.plot(rxs, So_uy / norm, "m-.", linewidth=2)
        if order % 2 == 1:
            ax1.plot(rxs, -So_ux / norm, "c:", linewidth=2)
            ax1.plot(rxs, -So_uy / norm, "m:", linewidth=2)

        # ax1.plot(rxs, abs(struc_func_ux[iorder])/abs(struc_func_uy[iorder]),
        #          'k', linewidth=1)

        ax1.plot(rxs, abs(So_ux) / abs(So_uy), "k", linewidth=1)

        # if self.orders[iorder]%2 == 1:
        #     ax1.plot(rxs, -struc_func_ux[iorder]/norm, '--b', linewidth=2)
        #     ax1.plot(rxs, -struc_func_uy[iorder]/norm, '--m', linewidth=2)

        cond = rxs < 6 * deltax
        ax1.plot(
            rxs[cond], 1.0e4 * rxs[cond] ** (order) / norm[cond], "k", linewidth=2
        )
        ax1.plot(rxs, rxs ** (order / 3) / norm, "--k", linewidth=2)

        ax1.plot(rxs, 1.0e0 * rxs ** (1) / norm, ":k", linewidth=2)

        z_bottom_axe = 0.09
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)

        ax2.set_xlabel("$r_x$")
        ax2.set_ylabel("flatness")
        ax2.set_xscale("log")
        ax2.set_yscale("log")

        # iorder4 = self.iorder_from_order(4)
        # iorder2 = self.iorder_from_order(2)

        # if delta_t != 0.:
        #     for it in range(imin_plot,imax_plot+1,delta_i_plot):
        #         struc_func_ux = dset_struc_func_ux[it]
        #         struc_func_ux = struc_func_ux.reshape(
        #             [self.norders, self.nrx])
        #         struc_func_uy = dset_struc_func_uy[it]
        #         struc_func_uy = struc_func_uy.reshape(
        #             [self.norders, self.nrx])

        #         flatnessL = struc_func_ux[iorder4]/struc_func_ux[iorder2]**2
        #         flatnessT = struc_func_uy[iorder4]/struc_func_uy[iorder2]**2

        #         ax2.plot(rxs, flatnessL, 'c', linewidth=1)
        #         ax2.plot(rxs, flatnessT, 'm', linewidth=1)

        # struc_func_ux = dset_struc_func_ux[imin_plot:imax_plot+1].mean(0)
        # struc_func_ux = struc_func_ux.reshape([self.norders, self.nrx])
        # struc_func_uy = dset_struc_func_uy[imin_plot:imax_plot+1].mean(0)
        # struc_func_uy = struc_func_uy.reshape([self.norders, self.nrx])

        # flatnessL = struc_func_ux[iorder4]/struc_func_ux[iorder2]**2
        # flatnessT = struc_func_uy[iorder4]/struc_func_uy[iorder2]**2
        # ax2.plot(rxs, flatnessL, 'c', linewidth=2)
        # ax2.plot(rxs, flatnessT, 'm', linewidth=2)

        S2_eta = self.strfunc_from_pdf(pdf_eta, values_inc_eta, 2)
        S2_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, 2)
        S2_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, 2)

        S4_eta = self.strfunc_from_pdf(pdf_eta, values_inc_eta, 4)
        S4_ux = self.strfunc_from_pdf(pdf_ux, values_inc_ux, 4)
        S4_uy = self.strfunc_from_pdf(pdf_uy, values_inc_uy, 4)

        flatnessL_bis = S4_ux / S2_ux ** 2
        flatnessT_bis = S4_uy / S2_uy ** 2
        flatness_eta = S4_eta / S2_eta ** 2

        ax2.plot(rxs, flatnessL_bis, "c--", linewidth=2)
        ax2.plot(rxs, flatnessT_bis, "m--", linewidth=2)
        ax2.plot(rxs, flatness_eta, "y--", linewidth=2)

        cond = np.logical_and(rxs < 70 * deltax, rxs > 5 * deltax)
        ax2.plot(rxs[cond], 1e1 * rxs[cond] ** (-1), ":k", linewidth=2)

        ax2.plot(rxs, 3 * np.ones(rxs.shape), "k--", linewidth=0.5)

    def plot_Kolmo(self, tmin=0, tmax=None):
        """Plot quantities appearing in the Kolmogorov law."""
        with h5py.File(self.path_file, "r") as h5file:
            times = h5file["times"][...]

            if tmax is None:
                tmax = times.max()

            rxs = h5file["rxs"][...]

            oper = h5file["/info_simul/params/oper"]
            nx = oper.attrs["nx"]
            Lx = oper.attrs["Lx"]
            deltax = Lx / nx

            rxs = np.array(rxs, dtype=np.float64) * deltax

            imin_plot = np.argmin(abs(times - tmin))
            imax_plot = np.argmin(abs(times - tmax))

            tmin_plot = times[imin_plot]
            tmax_plot = times[imax_plot]

            to_print = f"plot(tmin={tmin}, tmax={tmax})"
            print(to_print)

            to_print = """plot structure functions
    tmin = {:8.6g} ; tmax = {:8.6g}
    imin = {:8d} ; imax = {:8d}""".format(
                tmin_plot, tmax_plot, imin_plot, imax_plot
            )
            print(to_print)

            # dset_struc_func_ux = h5file['struc_func_ux']
            # struc_func_ux = dset_struc_func_ux[imin_plot:imax_plot+1].mean(0)
            # struc_func_ux = struc_func_ux.reshape([self.norders, self.nrx])
            # order = 3
            # iorder = self.iorder_from_order(order)
            # S_ux3 = struc_func_ux[iorder]

            S_uL2JL = h5file["struc_func_uL2JL"][imin_plot : imax_plot + 1].mean(
                0
            )
            S_uT2JL = h5file["struc_func_uT2JL"][imin_plot : imax_plot + 1].mean(
                0
            )
            S_c2h2uL = h5file["struc_func_c2h2uL"][
                imin_plot : imax_plot + 1
            ].mean(0)
            S_Kolmo = h5file["struc_func_Kolmo"][imin_plot : imax_plot + 1].mean(
                0
            )
            # S_uT2uL = h5file["struc_func_uT2uL"][imin_plot : imax_plot + 1].mean(
            #     0
            # )

        S_Kolmo_theo = -4 * rxs

        x_left_axe = 0.12
        z_bottom_axe = 0.56
        width_axe = 0.85
        height_axe = 0.37
        size_axe = [x_left_axe, z_bottom_axe, width_axe, height_axe]
        fig, ax1 = self.output.figure_axe(size_axe=size_axe)
        ax1.set_xlabel("$r_x$")
        ax1.set_title("struct. functions\n" + self.output.summary_simul)
        ax1.set_xscale("log")
        ax1.set_yscale("linear")

        ax1.set_ylabel("struct. functions")

        ax1.plot(rxs, S_Kolmo / S_Kolmo_theo, "k", linewidth=2)
        ax1.plot(rxs, (S_uL2JL + S_uT2JL) / S_Kolmo_theo, "r", linewidth=2)

        ax1.plot(rxs, S_c2h2uL / S_Kolmo_theo, "b", linewidth=2)

        ax1.plot(rxs, S_uL2JL / S_Kolmo_theo, "r--", linewidth=1)
        ax1.plot(rxs, S_uT2JL / S_Kolmo_theo, "r-.", linewidth=1)

        ax1.plot(
            rxs, (S_uL2JL + S_uT2JL + S_c2h2uL) / S_Kolmo_theo, "y", linewidth=1
        )

        cond = rxs < 6 * deltax
        ax1.plot(
            rxs[cond],
            1.0e0 * rxs[cond] ** 3 / S_Kolmo_theo[cond],
            "k",
            linewidth=2,
        )

        ax1.plot(rxs, np.ones(rxs.shape), "k:", linewidth=1)

        ax1.set_ylim([5e-2, 1.5])

        z_bottom_axe = 0.09
        size_axe[1] = z_bottom_axe
        ax2 = fig.add_axes(size_axe)

        ax2.set_xlabel("$r_x$")
        ax2.set_ylabel("ratio S_ux3/S_uT2uL")
        ax2.set_xscale("log")
        ax2.set_yscale("linear")

        # ax2.plot(rxs, S_ux3/S_uT2uL, 'k', linewidth=2)

        ax2.plot(rxs, S_uL2JL / S_uT2JL, "k--", linewidth=2)

        ax2.plot(rxs, 3 * np.ones(rxs.shape), "k:", linewidth=1)

        ax2.set_ylim([2, 5])



"""create new files with a modificated resolution."""

from solveq2d import solveq2d

t_approx = None
coef_modif_resol = 2
dir_base = "~/Storage/Results_SW1lw/Pure_standing_waves_1920x1920"

print("run the function solveq2d.modif_resolution_all_dir")

solveq2d.modif_resolution_all_dir(
    t_approx=t_approx, coef_modif_resol=coef_modif_resol, dir_base=dir_base
)



"""
Spatiotemporal Spectra (:mod:`fluidsim.base.output.spatiotemporal_spectra`)
===========================================================================

Provides:

.. autoclass:: SpatioTemporalSpectra3D
   :members:
   :private-members:

.. autoclass:: SpatioTemporalSpectra2D
   :members:
   :private-members:

"""

from pathlib import Path
from logging import warn
from math import pi

import numpy as np
from scipy import signal
import h5py
from rich.progress import Progress

from fluiddyn.util import mpi
from fluidsim.util.util import open_patient
from fluidsim.base.output.base import SpecificOutput

from transonic import boost, Array, Type

Uf32f64 = Type(np.float32, np.float64)
A = Array[Uf32f64, "1d"]


@boost
def find_index_first_geq(arr: A, value: Uf32f64):
    """find the first index such that `arr[index] >= value`"""
    for i, v in enumerate(arr):
        if v >= value:
            return i
    raise ValueError("No index such that `arr[index] >= value`")


@boost
def find_index_first_g(arr: A, value: Uf32f64):
    """find the first index such that `arr[index] > value`"""
    for i, v in enumerate(arr):
        if v > value:
            return i
    raise ValueError("No index such that `arr[index] > value`")


@boost
def find_index_first_l(arr: A, value: Uf32f64):
    """find the first index such that `arr[index] < value`"""
    for i, v in enumerate(arr):
        if v < value:
            return i
    raise ValueError("No index such that `arr[index] < value`")


def filter_tmins_paths(tmin, tmins, paths):
    if tmins.size == 1:
        return tmins, paths
    delta_tmin = np.diff(tmins).min()
    start = find_index_first_l(tmin - tmins, delta_tmin)
    return tmins[start:], paths[start:]


@boost
def get_arange_minmax(times: A, tmin: Uf32f64, tmax: Uf32f64):
    """get a range of index for which `tmin <= times[i] <= tmax`

    This assumes that `times` is sorted.

    """

    if tmin <= times[0]:
        start = 0
    else:
        start = find_index_first_geq(times, tmin)

    if tmax >= times[-1]:
        stop = len(times)
    else:
        stop = find_index_first_g(times, tmax)

    return np.arange(start, stop)


class SpatioTemporalSpectra3D(SpecificOutput):
    """
    Computes the spatiotemporal spectra.
    """

    _tag = "spatiotemporal_spectra"
    nb_dim = 3

    @classmethod
    def _complete_params_with_default(cls, params):
        params.output.periods_save._set_attrib(cls._tag, 0)
        params.output._set_child(
            cls._tag,
            attribs={
                "probes_region": None,
                "file_max_size": 10.0,  # MB
                "SAVE_AS_COMPLEX64": True,
            },
        )

        params.output.spatiotemporal_spectra._set_doc(
            """
            probes_region: int tuple (default:None)

                Boundaries of the region to record in the spectral domain.

                probes_region = (ikxmax, ikymax, ikzmax), in nondimensional units (mode indices).

                The resulting ranges for each wavevectors are :

                    ikx in [0 , ikxmax]

                    iky in [-ikymax+1 , ikymax]

                    ikz in [-ikzmax+1 , ikzmax]

                If None, set all ikmax = 4.

            file_max_size: float (default: 10.0)

                Maximum size of one time series file, in megabytes.

            SAVE_AS_COMPLEX64: bool (default: True)

                If set to False, probes data is saved as complex128.

                Warning : saving as complex128 reduces digital noise at high frequency, but doubles the size of the output!

            """
        )

    def __init__(self, output):
        params = output.sim.params
        try:
            params_st_spec = params.output.spatiotemporal_spectra
        except AttributeError:
            warn(
                "Cannot initialize spatiotemporal spectra output because "
                "`params` does not contain parameters for this class."
            )
            return

        super().__init__(
            output,
            period_save=params.output.periods_save.spatiotemporal_spectra,
        )

        oper = self.sim.oper

        # Parameters
        self.period_save = params.output.periods_save.spatiotemporal_spectra

        self.path_dir = Path(self.sim.output.path_run) / "spatiotemporal"
        self.keys_fields = self.sim.info_solver.classes.State.keys_state_phys

        if not output._has_to_save:
            self.period_save = 0.0
        if self.period_save == 0.0:
            return

        if params_st_spec.probes_region is not None:
            if self.nb_dim == 3:
                ikxmax, ikymax, ikzmax = params_st_spec.probes_region
            else:
                ikxmax, ikymax = params_st_spec.probes_region
        else:
            ikxmax = ikymax = 4
            if self.nb_dim == 3:
                ikzmax = 4

        ikxmax = min(int(round(ikxmax)), params.oper.nx // 2)
        ikymax = min(int(round(ikymax)), params.oper.ny // 2)

        if self.nb_dim == 3:
            ikzmax = min(int(round(ikzmax)), params.oper.nz // 2)
            self.probes_region = ikxmax, ikymax, ikzmax
        else:
            self.probes_region = ikxmax, ikymax

        self.file_max_size = params_st_spec.file_max_size
        self.SAVE_AS_COMPLEX64 = params_st_spec.SAVE_AS_COMPLEX64

        # region must be int tuple
        ikxmax = int(ikxmax)
        ikymax = int(ikymax)
        if self.nb_dim == 3:
            ikzmax = int(ikzmax)

        # dimensions order in Fourier space
        if self.nb_dim == 3:
            self.dims_order = np.array(oper.oper_fft.get_dimX_K())
        else:
            self.dims_order = np.arange(2)
            if oper.oper_fft.get_is_transposed():
                self.dims_order = self.dims_order[::-1]

        # data directory
        if mpi.rank == 0:
            self.path_dir.mkdir(exist_ok=True)
        if mpi.nb_proc > 1:
            mpi.comm.barrier()

        # data type and size
        if self.SAVE_AS_COMPLEX64:
            self.datatype = np.complex64
            size_1_number = 8e-6
        else:
            self.datatype = np.complex128
            size_1_number = 16e-6

        # check for existing files
        paths = sorted(self.path_dir.glob("rank*.h5"))
        if paths:
            # check values in files
            with open_patient(paths[0], "r") as file:
                if file.attrs["nb_proc"] != mpi.nb_proc:
                    raise ValueError(
                        f"process number ({mpi.nb_proc}) is different from "
                        f"process number in file ({file.attrs['nb_proc']})"
                    )
                if (file.attrs["dims_order"] != self.dims_order).any():
                    raise ValueError("dimensions order is different from files")
                if (file.attrs["probes_region"] != self.probes_region).any():
                    raise ValueError("probes region is different from files")
            # init from files
            INIT_FROM_PARAMS = False
            paths_rank = [
                p for p in paths if p.name.startswith(f"rank{mpi.rank:05}")
            ]
            if paths_rank:
                self.path_file = paths_rank[-1]
                with open_patient(self.path_file, "r") as file:
                    self.index_file = file.attrs["index_file"]
                    self.probes_k0adim_loc = file["probes_k0adim_loc"][:]
                    self.probes_ik0_loc = file["probes_ik0_loc"][:]
                    self.probes_k1adim_loc = file["probes_k1adim_loc"][:]
                    self.probes_ik1_loc = file["probes_ik1_loc"][:]

                    if self.nb_dim == 3:
                        self.probes_k2adim_loc = file["probes_k2adim_loc"][:]
                        self.probes_ik2_loc = file["probes_ik2_loc"][:]

                    self.probes_nb_loc = self.probes_ik0_loc.size
                    self.number_times_in_file = file["times"].size
                    self.t_last_save = file["times"][-1]
            else:
                # no probes in proc
                self.path_file = None
                self.index_file = 0
                self.number_times_in_file = 0
                self.probes_nb_loc = 0
                self.probes_ik0_loc = []
                self.probes_ik1_loc = []
                self.probes_ik2_loc = []

                with open_patient(paths[-1], "r") as file:
                    self.t_last_save = file["times"][-1]

        else:
            # no files were found : initialize from params
            INIT_FROM_PARAMS = True

            if self.nb_dim == 3:
                # pair kx,ky,kz with k0,k1,k2
                iksmax = np.array([ikzmax, ikymax, ikxmax])
                iksmin = np.array([1 - ikzmax, 1 - ikymax, 0])
                ik0max, ik1max, ik2max = iksmax[self.dims_order]
                ik0min, ik1min, ik2min = iksmin[self.dims_order]

                # local probes indices
                (
                    k0_adim_loc,
                    k1_adim_loc,
                    k2_adim_loc,
                ) = oper.oper_fft.get_k_adim_loc()
                K0_adim, K1_adim, K2_adim = np.meshgrid(
                    k0_adim_loc, k1_adim_loc, k2_adim_loc, indexing="ij"
                )
            else:
                iksmax = np.array([ikymax, ikxmax])
                iksmin = np.array([1 - ikymax, 0])
                ik0max, ik1max = iksmax[self.dims_order]
                ik0min, ik1min = iksmin[self.dims_order]

                kx_adim_loc = np.array(
                    (oper.kx_loc / oper.deltakx).round(), dtype=int
                )
                ky_adim_loc = np.array(
                    (oper.ky_loc / oper.deltaky).round(), dtype=int
                )
                if oper.oper_fft.get_is_transposed():
                    k0_adim_loc = kx_adim_loc
                    k1_adim_loc = ky_adim_loc
                else:
                    k0_adim_loc = ky_adim_loc
                    k1_adim_loc = kx_adim_loc

                K0_adim, K1_adim = np.meshgrid(
                    k0_adim_loc, k1_adim_loc, indexing="ij"
                )

            cond_region = (
                (K0_adim >= ik0min)
                & (K0_adim <= ik0max)
                & (K1_adim >= ik1min)
                & (K1_adim <= ik1max)
            )

            if self.nb_dim == 3:
                cond_region = (
                    cond_region & (K2_adim >= ik2min) & (K2_adim <= ik2max)
                )

            if self.nb_dim == 3:
                (
                    self.probes_ik0_loc,
                    self.probes_ik1_loc,
                    self.probes_ik2_loc,
                ) = np.where(cond_region)
            else:
                (
                    self.probes_ik0_loc,
                    self.probes_ik1_loc,
                ) = np.where(cond_region)

            self.probes_nb_loc = self.probes_ik0_loc.size

            # local probes wavenumbers (nondimensional)
            self.probes_k0adim_loc = self._get_data_probe_from_field(K0_adim)
            self.probes_k1adim_loc = self._get_data_probe_from_field(K1_adim)

            if self.nb_dim == 3:
                self.probes_k2adim_loc = self._get_data_probe_from_field(K2_adim)

            # initialize files info
            self.index_file = 0
            self.number_times_in_file = 0
            self.t_last_save = -self.period_save

        # size of a single write: nb_fields * probes_nb_loc + time
        probes_write_size = (
            len(self.keys_fields) * self.probes_nb_loc + 1
        ) * size_1_number
        self.max_number_times_in_file = int(
            self.file_max_size / probes_write_size
        )

        # initialize files
        if INIT_FROM_PARAMS and self.probes_nb_loc > 0:
            self._init_new_file(tmin_file=self.sim.time_stepping.t)

    def _init_files(self, arrays_1st_time=None):
        # we don't want to do anything when this function is called.
        pass

    def _init_new_file(self, tmin_file=None):
        """Initializes a new file"""
        if tmin_file is not None:
            # max number of digits = int(log10(t_end)) + 1
            # add .3f precision = 4 additional characters
            # +2 by anticipation of potential restarts
            str_width = int(np.log10(self.sim.params.time_stepping.t_end)) + 7
            ind_str = f"tmin{tmin_file:0{str_width}.3f}"
        else:
            ind_str = f"file{self.index_file:04}"
        self.path_file = self.path_dir / f"rank{mpi.rank:05}_{ind_str}.h5"
        with open_patient(self.path_file, "w") as file:
            file.attrs["nb_proc"] = mpi.nb_proc
            file.attrs["dims_order"] = self.dims_order
            file.attrs["index_file"] = self.index_file
            file.attrs["probes_region"] = self.probes_region
            file.attrs["period_save"] = self.period_save
            file.attrs["max_number_times_in_file"] = self.max_number_times_in_file
            create_ds = file.create_dataset
            create_ds("probes_k0adim_loc", data=self.probes_k0adim_loc)
            create_ds("probes_ik0_loc", data=self.probes_ik0_loc)
            create_ds("probes_k1adim_loc", data=self.probes_k1adim_loc)
            create_ds("probes_ik1_loc", data=self.probes_ik1_loc)
            if self.nb_dim == 3:
                create_ds("probes_k2adim_loc", data=self.probes_k2adim_loc)
                create_ds("probes_ik2_loc", data=self.probes_ik2_loc)
            for key in self.keys_fields:
                create_ds(
                    key + "_Fourier_loc",
                    (self.probes_nb_loc, 1),
                    maxshape=(self.probes_nb_loc, None),
                    dtype=self.datatype,
                )
            create_ds("times", (1,), maxshape=(None,))

    def _write_to_file(self, data):
        """Writes a file with the temporal data"""
        with open_patient(self.path_file, "a") as file:
            for k, v in data.items():
                dset = file[k]
                if k.startswith("times"):
                    dset.resize((self.number_times_in_file,))
                    dset[-1] = v
                else:
                    dset.resize((self.probes_nb_loc, self.number_times_in_file))
                    dset[:, -1] = v

    def _add_probes_data_to_dict(self, data, key):
        """Probes fields in Fourier space and append data to a dict object"""
        data[key + "_Fourier_loc"] = self._get_data_probe_from_field(
            self.sim.state.get_var(f"{key}_fft")
        )

    def _online_save(self):
        """Prepares data and writes to file"""
        tsim = self.sim.time_stepping.t
        if (
            tsim + 1e-15
        ) // self.period_save > self.t_last_save // self.period_save:
            # if max write number is reached, init new file
            if self.number_times_in_file >= self.max_number_times_in_file:
                self.index_file += 1
                self.number_times_in_file = 0
                self._init_new_file(tmin_file=self.sim.time_stepping.t)
            # get data from probes
            data = {"times": self.sim.time_stepping.t}
            for key in self.keys_fields:
                self._add_probes_data_to_dict(data, key)
            # write to file
            self.number_times_in_file += 1
            if self.probes_nb_loc > 0:
                self._write_to_file(data)
            self.t_last_save = tsim

    def load_time_series(self, keys=None, tmin=0, tmax=None, dtype=None):
        """load time series from files"""

        if mpi.nb_proc > 1:
            raise RuntimeError(
                "This postprocessing function should not be called with MPI."
            )

        if keys is None:
            keys = self.keys_fields
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        # get ranks
        paths = sorted(self.path_dir.glob("rank*.h5"))
        ranks = sorted({int(p.name[4:9]) for p in paths})

        # get times and dimensions order from the files of first rank
        print(f"load times series...")
        paths_1st_rank = [
            p for p in paths if p.name.startswith(f"rank{ranks[0]:05}")
        ]

        with open_patient(paths_1st_rank[0], "r") as file:
            dims_order = file.attrs["dims_order"]
            region = file.attrs["probes_region"]
            if dtype is None:
                dtype = file[keys[0] + "_Fourier_loc"].dtype

        # get list of useful files, from tmin
        tmins_files = np.array([float(p.name[14:-3]) for p in paths_1st_rank])
        tmins_files, paths_1st_rank = filter_tmins_paths(
            tmin, tmins_files, paths_1st_rank
        )

        with Progress() as progress:
            npaths = len(paths_1st_rank)
            task_files = progress.add_task(
                "Getting times from rank 0...", total=npaths
            )

            times = []
            for ip, path in enumerate(paths_1st_rank):
                with open_patient(path, "r") as file:
                    if tmins_files[ip] > tmax:
                        progress.update(task_files, completed=npaths)
                        break
                    times_file = file["times"][:]
                    cond_times = (times_file >= tmin) & (times_file <= tmax)
                    times.append(times_file[cond_times])
                    progress.update(task_files, advance=1)

        times = np.concatenate(times)

        tmin = times.min()
        tmax = times.max()
        print(f"tmin={tmin:8.6g}, tmax={tmax:8.6g}, nit={times.size}")

        # get sequential shape of Fourier space
        if self.nb_dim == 3:
            ikxmax, ikymax, ikzmax = region
            iksmax = np.array([ikzmax, ikymax, ikxmax])
            iksmin = np.array([1 - ikzmax, 1 - ikymax, 0])
            ik0max, ik1max, ik2max = iksmax[dims_order]
            ik0min, ik1min, ik2min = iksmin[dims_order]
            shape_series = (
                ik0max + 1 - ik0min,
                ik1max + 1 - ik1min,
                ik2max + 1 - ik2min,
                times.size,
            )
        else:
            ikxmax, ikymax = region
            iksmax = np.array([ikymax, ikxmax])
            iksmin = np.array([1 - ikymax, 0])
            ik0max, ik1max = iksmax[dims_order]
            ik0min, ik1min = iksmin[dims_order]
            shape_series = (
                ik0max + 1 - ik0min,
                ik1max + 1 - ik1min,
                times.size,
            )

        # load series, rebuild as state_spect arrays + time
        series = {
            f"{k}_Fourier": np.empty(shape_series, dtype=dtype) for k in keys
        }
        with Progress() as progress:
            task_ranks = progress.add_task("Rearranging...", total=len(ranks))
            task_files = progress.add_task("Rank 00000...", total=npaths)
            # loop on all files, rank by rank
            for rank in ranks:
                paths_rank = [
                    p for p in paths if p.name.startswith(f"rank{rank:05}")
                ]

                # get list of useful files, from tmin
                tmins_files = np.array([float(p.name[14:-3]) for p in paths_rank])
                tmins_files, paths_rank = filter_tmins_paths(
                    tmin, tmins_files, paths_rank
                )

                npaths = len(paths_rank)
                progress.update(
                    task_files,
                    description=f"Rank {rank:05}...",
                    total=npaths,
                    completed=0,
                )

                # for a given rank, paths are sorted by time
                for ip, path_file in enumerate(paths_rank):
                    # break after the last useful file
                    if tmins_files[ip] > tmax:
                        progress.update(task_files, completed=npaths)
                        break

                    with open_patient(path_file, "r") as file:
                        # time indices
                        times_file = file["times"][:]
                        its_file = get_arange_minmax(times_file, tmin, tmax)
                        tmin_keep = times_file[its_file[0]]
                        tmax_keep = times_file[its_file[-1]]
                        its = get_arange_minmax(times, tmin_keep, tmax_keep)

                        # k_adim_loc = global probes indices!
                        ik0 = file["probes_k0adim_loc"][:]
                        ik1 = file["probes_k1adim_loc"][:]
                        if self.nb_dim == 3:
                            ik2 = file["probes_k2adim_loc"][:]

                        # load data at desired times for all keys_fields
                        for key in keys:
                            skey = key + "_Fourier"
                            data = file[skey + "_loc"][:, its_file]

                            if self.nb_dim == 3:
                                for i, it in enumerate(its):
                                    series[skey][ik0, ik1, ik2, it] = data[:, i]
                            else:
                                for i, it in enumerate(its):
                                    series[skey][ik0, ik1, it] = data[:, i]

                    # update rich task
                    progress.update(task_files, advance=1)

                # update rich task
                progress.update(task_ranks, advance=1)

        # add Ki_adim arrays, times and dims order
        k0_adim = np.r_[0 : ik0max + 1, ik0min:0]
        k1_adim = np.r_[0 : ik1max + 1, ik1min:0]

        if self.nb_dim == 3:
            k2_adim = np.r_[0 : ik2max + 1, ik2min:0]
            K0_adim, K1_adim, K2_adim = np.meshgrid(
                k0_adim, k1_adim, k2_adim, indexing="ij"
            )
        else:
            K0_adim, K1_adim = np.meshgrid(k0_adim, k1_adim, indexing="ij")

        series.update(
            {
                "K0_adim": K0_adim,
                "K1_adim": K1_adim,
                "times": times,
                "dims_order": dims_order,
            }
        )
        if self.nb_dim == 3:
            series["K2_adim"] = K2_adim

        return series

    def _compute_spectrum(self, data):
        if not hasattr(self, "f_sample"):
            paths = sorted(self.path_dir.glob("rank*.h5"))
            with h5py.File(paths[0], "r") as file:
                self.f_sample = 1.0 / file.attrs["period_save"]
            self.domega = 2 * pi * self.f_sample / data.shape[-1]

        # TODO: I'm not sure if detrend=False is good in prod, but it's much
        # better for testing
        freq, spectrum = signal.periodogram(
            data,
            fs=self.f_sample,
            scaling="spectrum",
            detrend=False,
            return_onesided=False,
        )
        return freq, spectrum / self.domega

    def compute_spectra(self, tmin=0, tmax=None, dtype=None):
        """compute spatiotemporal spectra from files"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        # load time series as state_spect arrays + times
        series = self.load_time_series(tmin=tmin, tmax=tmax, dtype=dtype)

        # compute spectra
        print("performing time fft...")

        spectra = {k: v for k, v in series.items() if k.startswith("K")}

        for key, data in series.items():
            if "_Fourier" not in key:
                continue
            key_spectrum = "spectrum_" + key.split("_Fourier")[0]
            freq, spectrum = self._compute_spectrum(data)
            spectra[key_spectrum] = spectrum

        spectra["omegas"] = 2 * pi * freq
        spectra["dims_order"] = series["dims_order"]

        return spectra

    def _get_data_probe_from_field(self, field):
        return field[
            self.probes_ik0_loc,
            self.probes_ik1_loc,
            self.probes_ik2_loc,
        ]


class SpatioTemporalSpectra2D(SpatioTemporalSpectra3D):
    nb_dim = 2

    def _get_data_probe_from_field(self, field):
        return field[
            self.probes_ik0_loc,
            self.probes_ik1_loc,
        ]


class SpatioTemporalSpectraNS:
    def _get_path_saved_spectra(self, tmin, tmax, dtype, save_urud):
        base = f"periodogram_{tmin}_{tmax}"
        if dtype is not None:
            base += f"_{dtype}"
        if save_urud:
            base += "_urud"
        return self.path_dir / (base + ".h5")

    def _get_path_saved_tspectra(self, tmin, tmax, dtype, save_urud):
        base = f"periodogram_temporal_{tmin}_{tmax}"
        if dtype is not None:
            base += f"_{dtype}"
        if save_urud:
            base += "_urud"
        return self.path_dir / (base + ".h5")

    def save_spectra_kzkhomega(
        self, tmin=0, tmax=None, dtype=None, save_urud=False
    ):
        """
        save:
            - the spatiotemporal spectra, with a cylindrical average in k-space
            - the temporal spectra, with an average on the whole k-space
        """
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        # compute spectra
        print("Computing spectra...")
        spectra = self.compute_spectra(tmin=tmin, tmax=tmax, dtype=dtype)

        # get kz, kh
        params_oper = self.sim.params.oper
        deltakx = 2 * pi / params_oper.Lx
        order = spectra["dims_order"]
        KX = deltakx * spectra[f"K{order[-1]}_adim"]
        kx_max = self.sim.params.oper.nx // 2 * deltakx

        if self.nb_dim == 3:
            deltaky = 2 * pi / params_oper.Ly
            deltakz = 2 * pi / params_oper.Lz
            _deltakhs = [deltakx, deltaky]
            _ideltakh = np.argmax(_deltakhs)
            deltakh = _deltakhs[_ideltakh]
            KY = deltaky * spectra[f"K{order[1]}_adim"]
            KH = np.sqrt(KX ** 2 + KY ** 2)
            khmax_spectra = [KX, KY][_ideltakh].max()
            del KY
        else:
            # in 2d, vertical (here "z") is y
            deltakz = 2 * pi / params_oper.Ly
            KH = abs(KX)
            deltakh = deltakx
            khmax_spectra = KX.max()

        KZ = deltakz * spectra[f"K{order[0]}_adim"]

        kz_spectra = np.arange(0, KZ.max() + 1e-15, deltakz)

        nkh_spectra = max(2, int(khmax_spectra / deltakh))
        kh_spectra = deltakh * np.arange(nkh_spectra)

        # get one-sided frequencies
        omegas = spectra["omegas"]
        nomegas = (omegas.size + 1) // 2
        omegas_onesided = abs(omegas[:nomegas])

        # kzkhomega : perform cylindrical average
        # temporal spectra : average on Fourier space
        spectra_kzkhomega = {
            "kz_spectra": kz_spectra,
            "kh_spectra": kh_spectra,
            "omegas": omegas_onesided,
        }
        tspectra = {"omegas": omegas_onesided}
        for key, data in spectra.items():
            if not key.startswith("spectrum_"):
                continue
            spectra_kzkhomega[key] = self.compute_spectrum_kzkhomega(
                np.ascontiguousarray(data), kh_spectra, kz_spectra, KX, KZ, KH
            )
            tspectrum = self._sum_wavenumber(data, KX, kx_max)
            # one-sided frequencies
            tspectrum_onesided = np.zeros(nomegas)
            tspectrum_onesided[0] = tspectrum[0]
            tspectrum_onesided[1:] = (
                tspectrum[1:nomegas] + tspectrum[-1:-nomegas:-1]
            )
            tspectra[key] = tspectrum_onesided

        del spectra

        # total kinetic energy
        if self.nb_dim == 3:
            spectra_kzkhomega["spectrum_K"] = 0.5 * (
                spectra_kzkhomega["spectrum_vx"]
                + spectra_kzkhomega["spectrum_vy"]
                + spectra_kzkhomega["spectrum_vz"]
            )
            tspectra["spectrum_K"] = 0.5 * (
                tspectra["spectrum_vx"]
                + tspectra["spectrum_vy"]
                + tspectra["spectrum_vz"]
            )
        else:
            spectra_kzkhomega["spectrum_K"] = 0.5 * (
                spectra_kzkhomega["spectrum_ux"]
                + spectra_kzkhomega["spectrum_uy"]
            )
            tspectra["spectrum_K"] = 0.5 * (
                tspectra["spectrum_ux"] + tspectra["spectrum_uy"]
            )

        # potential energy
        try:
            N = self.sim.params.N
            spectra_kzkhomega["spectrum_A"] = (
                0.5 / N ** 2 * spectra_kzkhomega["spectrum_b"]
            )
            tspectra["spectrum_A"] = 0.5 / N ** 2 * tspectra["spectrum_b"]
        except AttributeError:
            pass

        # save to files
        path_file = self._get_path_saved_spectra(tmin, tmax, dtype, save_urud)
        with h5py.File(path_file, "w") as file:
            file.attrs["tmin"] = tmin
            file.attrs["tmax"] = tmax
            for key, val in spectra_kzkhomega.items():
                file.create_dataset(key, data=val)

        path_file = self._get_path_saved_tspectra(tmin, tmax, dtype, save_urud)
        with h5py.File(path_file, "w") as file:
            file.attrs["tmin"] = tmin
            file.attrs["tmax"] = tmax
            for key, val in tspectra.items():
                file.create_dataset(key, data=val)

        # toroidal/poloidal decomposition
        if save_urud:
            print("Computing ur, ud spectra...")
            spectra_urud_kzkhomega = {}
            tspectra_urud = {}
            spectra = self.compute_spectra_urud(tmin=tmin, tmax=tmax, dtype=dtype)

            for key, data in spectra.items():
                if not key.startswith("spectrum_"):
                    continue
                spectra_urud_kzkhomega[key] = self.compute_spectrum_kzkhomega(
                    np.ascontiguousarray(data), kh_spectra, kz_spectra, KX, KZ, KH
                )
                spectra_kzkhomega[key] = spectra_urud_kzkhomega[key]
                tspectrum = self._sum_wavenumber(data, KX, kx_max)
                # one-sided frequencies
                tspectrum_onesided = np.zeros(nomegas)
                tspectrum_onesided[0] = tspectrum[0]
                tspectrum_onesided[1:] = (
                    tspectrum[1:nomegas] + tspectrum[-1:-nomegas:-1]
                )
                tspectra_urud[key] = tspectrum_onesided
                tspectra[key] = tspectra_urud[key]

            path_file = self._get_path_saved_spectra(tmin, tmax, dtype, save_urud)
            with h5py.File(path_file, "a") as file:
                for key, val in spectra_urud_kzkhomega.items():
                    file.create_dataset(key, data=val)

            path_file = self._get_path_saved_tspectra(
                tmin, tmax, dtype, save_urud
            )
            with h5py.File(path_file, "a") as file:
                for key, val in tspectra_urud.items():
                    file.create_dataset(key, data=val)

        return spectra_kzkhomega, tspectra

    def load_spectra_kzkhomega(
        self, tmin=0, tmax=None, dtype=None, save_urud=False
    ):
        """load kzkhomega spectra from file"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        spectra = {}

        path_file = self._get_path_saved_spectra(tmin, tmax, dtype, save_urud)
        with h5py.File(path_file, "r") as file:
            for key in file.keys():
                spectra[key] = file[key][...]

        return spectra

    def plot_kzkhomega(
        self,
        key_field=None,
        tmin=0,
        tmax=None,
        dtype=None,
        equation=None,
        xmax=None,
        ymax=None,
        cmap=None,
        vmin=None,
        vmax=None,
    ):
        """plot the spatiotemporal spectra, with a cylindrical average in k-space

        equation must start with 'omega=', 'kh=', 'kz=', 'ikh=' or 'ikz='.

        """
        keys_plot = self.keys_fields.copy()
        if self.nb_dim == 3:
            keys_plot.extend(["Khd", "Khr", "Kp"])

        if key_field is None:
            key_field = keys_plot[0]
        if key_field not in keys_plot:
            raise KeyError(f"possible keys are {keys_plot}")
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        key_spect = "spectrum_" + key_field
        if key_field.startswith("Kh") or key_field.startswith("Kp"):
            save_urud = True
        else:
            save_urud = False

        path_file = self._get_path_saved_spectra(tmin, tmax, dtype, save_urud)
        path_urud = self._get_path_saved_spectra(tmin, tmax, dtype, True)
        if path_urud.exists() and not path_file.exists():
            path_file = path_urud

        spectra_kzkhomega = {}

        # compute and save spectra if needed
        if not path_file.exists():
            if self.nb_dim == 3:
                self.save_spectra_kzkhomega(
                    tmin=tmin, tmax=tmax, dtype=dtype, save_urud=save_urud
                )
            else:
                self.save_spectra_kzkhomega(tmin=tmin, tmax=tmax, dtype=dtype)

        # load spectrum
        with h5py.File(path_file, "r") as file:
            if key_spect.startswith("spectrum_Kp"):
                spectrum = file["spectrum_Khd"][:] + 0.5 * file["spectrum_vz"][:]
            else:
                spectrum = file[key_spect][...]
            if dtype == "complex64":
                float_dtype = "float32"
            elif dtype == "complex128":
                float_dtype = "float64"
            if dtype:
                spectrum = spectrum.astype(float_dtype)
            spectra_kzkhomega[key_spect] = spectrum
            spectra_kzkhomega["kh_spectra"] = file["kh_spectra"][...]
            spectra_kzkhomega["kz_spectra"] = file["kz_spectra"][...]
            spectra_kzkhomega["omegas"] = file["omegas"][...]

        # slice along equation
        if equation is None:
            equation = f"omega=0"
        elif equation.startswith("kh="):
            kh = eval(equation[len("kh=") :])
            kh_spectra = spectra_kzkhomega["kh_spectra"]
            ikh = abs(kh_spectra - kh).argmin()
            equation = f"ikh={ikh}"
        elif equation.startswith("kz="):
            kz = eval(equation[len("kz=") :])
            kz_spectra = spectra_kzkhomega["kz_spectra"]
            ikz = abs(kz_spectra - kz).argmin()
            equation = f"ikz={ikz}"

        if equation.startswith("omega="):
            try:
                variables = dict(N=self.sim.params.N)
            except AttributeError:
                variables = dict()
            omega = eval(equation[len("omega=") :], variables)
            omegas = spectra_kzkhomega["omegas"]
            iomega = abs(omegas - omega).argmin()
            spect = spectra_kzkhomega[key_spect][:, :, iomega]
            xaxis = np.arange(spectra_kzkhomega["kh_spectra"].size)
            yaxis = np.arange(spectra_kzkhomega["kz_spectra"].size)
            xlabel = r"$k_h/\delta k_h$"
            ylabel = r"$k_z/\delta k_z$"
            omega = omegas[iomega]
            equation = r"$\omega=$" + f"{omega:.2g}"
            # use reduced frequency for stratified fluids
            try:
                N = self.sim.params.N
                equation = r"$\omega/N=$" + f"{omega/N:.2g}"
            except AttributeError:
                pass
        elif equation.startswith("ikh="):
            ikh = eval(equation[len("ikh=") :])
            kh_spectra = spectra_kzkhomega["kh_spectra"]
            spect = spectra_kzkhomega[key_spect][:, ikh, :].transpose()

            xaxis = np.arange(spectra_kzkhomega["kz_spectra"].size)
            yaxis = spectra_kzkhomega["omegas"]
            # use reduced frequency for stratified fluids
            try:
                N = self.sim.params.N
                yaxis /= N
            except AttributeError:
                pass

            xlabel = r"$k_z/\delta k_z$"
            ylabel = r"$\omega/N$"
            kh = kh_spectra[ikh]
            equation = f"$k_h = {ikh}\\delta k_h = {kh:.2g}$"
        elif equation.startswith("ikz="):
            ikz = eval(equation[len("ikz=") :])
            kz_spectra = spectra_kzkhomega["kz_spectra"]
            spect = spectra_kzkhomega[key_spect][ikz, :, :].transpose()

            xaxis = np.arange(spectra_kzkhomega["kh_spectra"].size)
            yaxis = spectra_kzkhomega["omegas"]
            # use reduced frequency for stratified fluids
            try:
                N = self.sim.params.N
                yaxis /= N
            except AttributeError:
                pass

            xlabel = r"$k_h/\delta k_h$"
            ylabel = r"$\omega/N$"
            kz = kz_spectra[ikz]
            equation = f"$k_z = {ikz}\\delta k_z = {kz:.2g}$"
        else:
            raise NotImplementedError(
                "equation must start with 'omega=', 'kh=', 'kz=', 'ikh=' or 'ikz='"
            )

        # plot
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

        # no log(0)
        spect += 1e-15

        im = ax.pcolormesh(
            xaxis,
            yaxis,
            np.log10(spect),
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
            shading="nearest",
        )
        fig.colorbar(im)

        ax.set_title(
            f"{key_field} spatiotemporal spectra {equation}\n"
            f"tmin={tmin:.3f}, tmax={tmax:.3f}\n" + self.output.summary_simul
        )

        # add dispersion relation : omega = N * kh / sqrt(kh ** 2 + kz ** 2)
        try:
            N = self.sim.params.N
        except AttributeError:
            return
        dkz_over_dkh = (
            spectra_kzkhomega["kz_spectra"][1]
            / spectra_kzkhomega["kh_spectra"][1]
        )
        if equation.startswith(r"$\omega"):
            if omega > 0 and omega <= N:
                ikz_disp = np.sqrt(N ** 2 / omega ** 2 - 1) / dkz_over_dkh * xaxis
                ax.plot(xaxis, ikz_disp, "k+", linewidth=2)
        elif equation.startswith(r"$k_h"):
            omega_disp = ikh / np.sqrt(ikh ** 2 + dkz_over_dkh ** 2 * xaxis ** 2)
            ax.plot(xaxis, omega_disp, "k+", linewidth=2)
        elif equation.startswith(r"$k_z"):
            omega_disp = xaxis / np.sqrt(
                xaxis ** 2 + dkz_over_dkh ** 2 * ikz ** 2
            )
            ax.plot(xaxis, omega_disp, "k+", linewidth=2)
        else:
            raise ValueError("wrong equation for dispersion relation")

        # set axis limits after plotting dispersion relation
        if xmax is None:
            xmax = xaxis.max()
        if ymax is None:
            ymax = yaxis.max()
        ax.set_xlim((0, xmax))
        ax.set_ylim((0, ymax))

    def compute_spectra_urud(self, tmin=0, tmax=None, dtype=None):
        raise NotImplementedError

    def compute_temporal_spectra(
        self, tmin=0, tmax=None, dtype=None, compute_urud=False
    ):
        """compute the temporal spectra by averaging over Fourier space"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        tspectra = {}

        # compute kxkykzomega spectra
        spectra = self.compute_spectra(tmin=tmin, tmax=tmax, dtype=dtype)
        if compute_urud:
            spectra.update(
                self.compute_spectra_urud(tmin=tmin, tmax=tmax, dtype=dtype)
            )

        # one-sided frequencies
        nomegas = (spectra["omegas"].size + 1) // 2
        tspectra["omegas"] = spectra["omegas"][:nomegas]

        order = spectra["dims_order"]
        KX = spectra[f"K{order[-1]}_adim"]
        deltakx = 2 * pi / self.sim.params.oper.Lx
        kx_max = self.sim.params.oper.nx // 2 * deltakx

        # average over Fourier space (kx,ky,kz)
        for key, spectrum in spectra.items():
            if not key.startswith("spectrum_"):
                continue
            tspectrum = self._sum_wavenumber(spectrum, KX, kx_max)
            # one-sided frequencies
            tspectrum_onesided = np.zeros(nomegas)
            tspectrum_onesided[0] = tspectrum[0]
            tspectrum_onesided[1:] = (
                tspectrum[1:nomegas] + tspectrum[-1:-nomegas:-1]
            )
            tspectra[key] = tspectrum_onesided

        # total kinetic energy
        if self.nb_dim == 3:
            tspectra["spectrum_K"] = 0.5 * (
                tspectra["spectrum_vx"]
                + tspectra["spectrum_vy"]
                + tspectra["spectrum_vz"]
            )
        else:
            tspectra["spectrum_K"] = 0.5 * (
                tspectra["spectrum_ux"] + tspectra["spectrum_uy"]
            )

        # potential energy
        try:
            N = self.sim.params.N
            tspectra["spectrum_A"] = 0.5 / N ** 2 * tspectra["spectrum_b"]
        except AttributeError:
            pass

        return tspectra

    def plot_temporal_spectra(
        self,
        key_field=None,
        tmin=0,
        tmax=None,
        dtype=None,
        xscale="log",
    ):
        """plot the temporal spectra computed from the 4d spectra"""
        keys_plot = self.keys_fields.copy()
        if self.nb_dim == 3:
            keys_plot.extend(["Khd", "Khr", "Kp"])
        if key_field is None:
            key_field = keys_plot[0]
        if key_field not in keys_plot:
            raise KeyError(f"possible keys are {keys_plot}")
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        if self.nb_dim == 3:
            # much simpler for 3d
            save_urud = True
        else:
            save_urud = False
        path_file = self._get_path_saved_tspectra(tmin, tmax, dtype, save_urud)
        if path_file.exists():
            tspectra = self.load_temporal_spectra(
                tmin=tmin, tmax=tmax, dtype=dtype, save_urud=save_urud
            )
        else:
            tspectra = self.save_temporal_spectra(
                tmin=tmin, tmax=tmax, dtype=dtype, save_urud=save_urud
            )

        # plot
        fig, ax = self.output.figure_axe()
        ax.set_xlabel(r"$\omega$")
        ax.set_ylabel("spectrum")
        ax.set_title(
            f"{key_field} temporal spectrum (tmin={tmin:.3f}, tmax={tmax:.3f})\n"
            + self.output.summary_simul
        )
        ax.set_xscale(xscale)
        ax.set_yscale("log")

        # specific to strat
        try:
            N = self.sim.params.N
        except AttributeError:
            ax.plot(
                tspectra["omegas"],
                tspectra["spectrum_" + key_field],
                "k",
                linewidth=2,
            )
        else:
            omegas = tspectra["omegas"] / N
            if self.nb_dim == 3:
                # polo/toro/potential decomposition
                EKp = tspectra["spectrum_Khd"] + 0.5 * tspectra["spectrum_vz"]
                EKhr = tspectra["spectrum_Khr"]
                EK = EKhr + EKp
                ax.plot(omegas, EK, "r", linewidth=2, label=r"$E_K$")
                ax.plot(omegas, EKhr, "r--", linewidth=1, label=r"$E_{K,toro}$")
                ax.plot(omegas, EKp, "r-.", linewidth=1, label=r"$E_{K,polo}$")
            else:
                # kinetic energy
                EK = tspectra["spectrum_K"]
                ax.plot(omegas, EK, "r", linewidth=2, label=r"$E_K$")
            EK_N = EK[abs(omegas - 1).argmin()]  # value at N
            EA = tspectra["spectrum_A"]

            ax.plot(omegas, EA, "b", linewidth=2, label=r"$E_A$")
            ax.set_title(
                f"kinetic/potential energy spectrum (tmin={tmin:.3f}, tmax={tmax:.3f})\n"
                + self.output.summary_simul
            )

            # resonant modes
            if self.nb_dim == 3:
                aspect_ratio = self.sim.oper.Lx / self.sim.oper.Lz
            else:
                aspect_ratio = self.sim.oper.Lx / self.sim.oper.Ly

            def modes(nx, nz):
                return np.sqrt(nx ** 2 / (nx ** 2 + aspect_ratio ** 2 * nz ** 2))

            nxs = np.arange(1, 11)
            modes_nz1 = modes(nxs, 1)
            modes_nz2 = modes(nxs, 2)
            modes_y = np.full_like(modes_nz1, fill_value=10 * EK_N)

            ax.plot(modes_nz1, modes_y, "o", label="modes $n_z=1$")
            ax.plot(modes_nz2, modes_y * 3, "o", label="modes $n_z=2$")

            # omega^-2 scaling
            omegas_scaling = np.arange(0.4, 1 + 1e-15, 0.01)
            scaling_y = EK_N * omegas_scaling ** -2

            ax.plot(omegas_scaling, scaling_y, "k--")

            # eye guide at N
            ax.axvline(1, linestyle="dotted")

            # eye guide at omega_f (specific to some forcing types)
            forcing_type = self.sim.params.forcing.type
            if forcing_type in ["watu_coriolis", "milestone"]:
                if forcing_type == "watu_coriolis":
                    omega_f = self.sim.params.forcing.watu_coriolis.omega_f
                elif forcing_type == "milestone":
                    period = self.sim.forcing.get_info()["period"]
                    omega_f = 2 * pi / period
                ax.axvline(omega_f / N, linestyle="dotted")

            ax.set_xlabel(r"$\omega/N$")

            ax.legend()

    def load_temporal_spectra(
        self, tmin=0, tmax=None, dtype=None, save_urud=False
    ):
        """load temporal spectra from file"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        tspectra = {}

        path_file = self._get_path_saved_tspectra(tmin, tmax, dtype, save_urud)
        with h5py.File(path_file, "r") as file:
            for key in file.keys():
                tspectra[key] = file[key][...]

        return tspectra

    def save_temporal_spectra(
        self, tmin=0, tmax=None, dtype=None, save_urud=False
    ):
        """compute temporal spectra from files"""
        if tmax is None:
            tmax = self.sim.params.time_stepping.t_end

        tspectra = self.compute_temporal_spectra(
            tmin=tmin, tmax=tmax, dtype=dtype, compute_urud=save_urud
        )

        path_file = self._get_path_saved_tspectra(tmin, tmax, dtype, save_urud)
        with h5py.File(path_file, "w") as file:
            file.attrs["tmin"] = tmin
            file.attrs["tmax"] = tmax
            for key, val in tspectra.items():
                file.create_dataset(key, data=val)

        return tspectra



import unittest

import numpy as np

from fluidsim.base.solvers.pseudo_spect import SimulBasePseudoSpectral

from fluidsim.util.testing import TestSimul, skip_if_no_fluidfft


@skip_if_no_fluidfft
class TestPreprocessPS(TestSimul):

    Simul = SimulBasePseudoSpectral

    @classmethod
    def init_params(cls):

        cls.params = params = SimulBasePseudoSpectral.create_default_params()
        params.short_name_type_run = "test_preprocess_ps"
        nh = 16
        Lh = 2 * np.pi
        params.oper.nx = nh
        params.oper.ny = nh
        params.oper.Lx = Lh
        params.oper.Ly = Lh
        params.init_fields.type = "constant"
        params.preprocess.enable = True
        params.preprocess.viscosity_type = "laplacian"
        params.preprocess.viscosity_scale = "energy"

    def test_set_viscosity(self):
        """Tests if preprocess can initialize viscosity or not."""
        sim = self.sim
        self.assertGreater(sim.params.nu_2, 0)
        self.assertEqual(sim.params.nu_4 + sim.params.nu_8, 0)
        assert np.any(sim.time_stepping.freq_lin > 0)
        sim.output.close_files()


if __name__ == "__main__":
    unittest.main()



"""Simple text output (:mod:`fluidsim.solvers.ns2d.output.print_stdout`)
========================================================================

.. autoclass:: PrintStdOutNS2D
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.output.print_stdout import PrintStdOutBase

from fluiddyn.util import mpi


class PrintStdOutNS2D(PrintStdOutBase):
    """Simple text output.

    Used to print in both the stdout and the stdout.txt file, and also
    to print simple info on the current state of the simulation.

    """

    def _make_str_info(self):
        to_print = super()._make_str_info()

        energy = self.output.compute_energy()

        if hasattr(self, "energy_tmp"):
            delta_energy = energy - self.energy_tmp
        else:
            delta_energy = 0.0

        if mpi.rank == 0:
            to_print += (
                f"              energy = {energy:9.3e} ; "
                f"Delta energy = {delta_energy:+9.3e}\n"
            )

            duration_left = self._evaluate_duration_left()
            if duration_left is not None:
                to_print += f"              estimated remaining duration = {duration_left}"

        self.energy_tmp = energy
        return to_print

    def load(self):
        dict_results = {"name_solver": self.output.name_solver}
        with open(self.output.path_run + "/stdout.txt") as file_means:
            lines = file_means.readlines()

        lines_t = []
        lines_E = []
        for il, line in enumerate(lines):
            if line[0:4] == "it =":
                lines_t.append(line)
            if line[0:22] == "              energy =":
                lines_E.append(line)

        nt = len(lines_t)
        if nt > 1:
            nt -= 1

        it = np.zeros(nt, dtype=int)
        t = np.zeros(nt)
        deltat = np.zeros(nt)

        E = np.zeros(nt)
        deltaE = np.zeros(nt)

        for il in range(nt):
            line = lines_t[il]
            words = line.split()
            it[il] = int(words[2])
            t[il] = float(words[6])
            deltat[il] = float(words[10])

            line = lines_E[il]
            words = line.split()
            E[il] = float(words[2])
            deltaE[il] = float(words[7])

        dict_results["it"] = it
        dict_results["t"] = t
        dict_results["deltat"] = deltat
        dict_results["E"] = E
        dict_results["deltaE"] = deltaE

        return dict_results

    def plot_deltat(self):
        dict_results = self.load()

        t = dict_results["t"]
        deltat = dict_results["deltat"]

        fig, ax = self.output.figure_axe()
        ax.set_xlabel("t")
        ax.set_ylabel("deltat(t)")

        ax.set_title("info stdout\n" + self.output.summary_simul)
        ax.plot(t, deltat, "k", linewidth=2)
        fig.tight_layout()

    def plot_energy(self):
        dict_results = self.load()

        t = dict_results["t"]
        E = dict_results["E"]
        deltaE = dict_results["deltaE"]

        fig, ax = self.output.figure_axe()

        ax.set_title("info stdout\n" + self.output.summary_simul)

        ax.set_xlabel("t")
        ax.set_ylabel("E(t), deltaE(t)")
        ax.plot(t, E, "k", linewidth=2)
        ax.plot(t, deltaE, "b", linewidth=2)
        fig.tight_layout()



import unittest
import numpy as np
import sys

import fluiddyn.util.mpi as mpi
from fluiddyn.util.paramcontainer import ParamContainer

from fluidsim.util.testing import TestCase, stdout_redirected, skip_if_no_fluidfft


def create_oper(type_fft=None, coef_dealiasing=2.0 / 3):

    params = ParamContainer(tag="params")

    params._set_attrib("ONLY_COARSE_OPER", False)
    params._set_attrib("f", 0)
    params._set_attrib("c2", 100)
    params._set_attrib("kd2", 0)

    from fluidsim.solvers.sw1l.operators import OperatorsPseudoSpectralSW1L

    OperatorsPseudoSpectralSW1L._complete_params_with_default(params)

    if mpi.nb_proc == 1:
        nh = 9
    else:
        nh = 8

    params.oper.nx = nh
    params.oper.ny = nh
    Lh = 6.0
    params.oper.Lx = Lh
    params.oper.Ly = Lh

    if type_fft is not None:
        params.oper.type_fft = type_fft

    params.oper.coef_dealiasing = coef_dealiasing

    oper = OperatorsPseudoSpectralSW1L(params=params)

    return oper


@skip_if_no_fluidfft
@unittest.skipIf(sys.platform.startswith("win"), "Untested on Windows")
class TestOperators(TestCase):
    @classmethod
    def setUpClass(cls):
        with stdout_redirected(cls.has_to_redirect_stdout):
            cls.oper = create_oper()
        cls.rtol = 1e-15
        cls.atol = 1e-14  # Absolute tolerance for double precision FFT

    def test_uxuyeta_qapam_conversion(self):
        """Test conversion back and forth from q,ap,am -> ux, uy, eta"""
        oper = self.oper
        q_fft = oper.create_arrayK_random()
        ap_fft = oper.create_arrayK_random()
        am_fft = oper.create_arrayK_random()
        if mpi.rank == 0:
            q_fft[0, 0] = ap_fft[0, 0] = am_fft[0, 0] = 0.0

        ux_fft, uy_fft, eta_fft = oper.uxuyetafft_from_qapamfft(
            q_fft, ap_fft, am_fft
        )
        q2_fft, ap2_fft, am2_fft = oper.qapamfft_from_uxuyetafft(
            ux_fft, uy_fft, eta_fft
        )

        np.testing.assert_allclose(q2_fft, q_fft, self.rtol, self.atol)
        np.testing.assert_allclose(ap2_fft, ap_fft, self.rtol, self.atol)
        np.testing.assert_allclose(am2_fft, am_fft, self.rtol, self.atol)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

"""Plate2d solver (:mod:`fluidsim.solvers.plate2d.solver`)
==========================================================

This module provides two classes defining the pseudo-spectral solver
plate2d.

.. autoclass:: InfoSolverPseudoSpectral
   :members:
   :private-members:

.. autoclass:: Simul
   :members:
   :private-members:

"""

import numpy as np

from fluidsim.base.setofvariables import SetOfVariables
from fluidsim.base.solvers.pseudo_spect import (
    SimulBasePseudoSpectral,
    InfoSolverPseudoSpectral,
)


class InfoSolverPlate2D(InfoSolverPseudoSpectral):
    """Contain the information on the solver plate2d."""

    def _init_root(self):
        """Init. `self` by writting the information on the solver.

        The function `InfoSolverPseudoSpectral._init_root` is
        called. We keep one class listed by this function:

        - :class:`fluidsim.base.time_stepping.pseudo_spect.TimeSteppingPseudoSpectral`

        The other first-level classes for this solver are:

        - :class:`fluidsim.solvers.plate2d.operators.OperatorsPseudoSpectralPlate2D`

        - :class:`fluidsim.solvers.plate2d.solver.Simul`

        - :class:`fluidsim.solvers.plate2d.state.StatePlate2D`

        - :class:`fluidsim.solvers.plate2d.init_fields.InitFieldsPlate2D`

        - :class:`fluidsim.solvers.plate2d.output.Output`

        - :class:`fluidsim.solvers.plate2d.forcing.ForcingPlate2D`

        """

        super()._init_root()

        package = "fluidsim.solvers.plate2d"
        self.module_name = package + ".solver"
        self.class_name = "Simul"
        self.short_name = "Plate2D"

        classes = self.classes

        classes.Operators.module_name = package + ".operators"
        classes.Operators.class_name = "OperatorsPseudoSpectralPlate2D"

        classes.State.module_name = package + ".state"
        classes.State.class_name = "StatePlate2D"

        classes.InitFields.module_name = package + ".init_fields"
        classes.InitFields.class_name = "InitFieldsPlate2D"

        classes.Output.module_name = package + ".output"
        classes.Output.class_name = "Output"

        classes.Forcing.module_name = package + ".forcing"
        classes.Forcing.class_name = "ForcingPlate2D"


class Simul(SimulBasePseudoSpectral):
    r"""Pseudo-spectral solver solving the Föppl-von Kármán equations.

    Notes
    -----

    .. |p| mathmacro:: \partial

    This class is dedicated to solve with a pseudo-spectral method the
    Föppl-von Kármán equations which describe the dynamics of a rigid
    plate.  Using the non-dimensional variables displacement :math:`z`
    and out of plane velocity :math:`w`:

    .. math::
       \p_t z = w,

    .. math::
       \p_t w = - \Delta^2 z + N_w(z) + f_w - \nu_\alpha (-\Delta)^\alpha w.

    where :math:`\Delta = \p_{xx} + \p_{yy}` is the Laplacian. The
    first term of the two equations corresponds to the linear part.
    :math:`f_w` and :math:`\nu_\alpha \Delta^\alpha w` are the
    forcing and the dissipation terms, respectively. The nonlinear
    term is equal to :math:`N_w(z) = \{ z, \chi \}`, where :math:`\{
    \cdot, \cdot \}` is the Monge-Ampère operator

    .. math::
       \{ a, b \} = \p_{xx} a \p_{yy} b + \p_{yy} a \p_{xx} b
       - 2 \p_{xy} a \p_{xy} b,

    and

    .. math:: \Delta^2 \chi = -\{ z, z \}.

    Taking the Fourier transform, we get:

    .. math::
       \p_t \hat z = \hat w,

    .. math::
       \p_t \hat w = - k^4 \hat z + \widehat{N_w(z)} + \hat f_w
       - \nu_\alpha k^{2\alpha} \hat w,

    where :math:`k^2 = |\mathbf{k}|^2`. For this simple solver, we
    will use the variables :math:`z` and :math:`w` and only the
    dissipative term will be solve exactly.  Thus, all the other terms
    are included in the :func:`tendencies_nonlin` function.

    **Energetics**: The total energy can be decomposed in the kinetic energy

    .. math::
       E_K = \frac{1}{2} \langle w^2 \rangle
       = \frac{1}{2} \sum_\mathbf{k} |\hat w|^2,

    the flexion energy

    .. math::
       E_L = \frac{1}{2} \langle (\Delta z)^2 \rangle
       = \frac{1}{2} \sum_\mathbf{k} k^4|\hat z|^2,

    and the non-quadratic extension energy

    .. math::
       E_E = \frac{1}{4} \langle (\Delta \chi)^2 \rangle
       = \frac{1}{4} \sum_\mathbf{k} k^4 |\hat \chi|^2.

    The energy injected into the system by the forcing is

    .. math::
       P = \langle f_w w \rangle,

    and the dissipation is

    .. math::
       D = \nu_\alpha \langle w (-\Delta)^\alpha w \rangle.

    """

    InfoSolver = InfoSolverPlate2D

    @staticmethod
    def _complete_params_with_default(params):
        """Complete the `params` container (static method)."""
        SimulBasePseudoSpectral._complete_params_with_default(params)
        attribs = {"beta": 0.0}
        params._set_attribs(attribs)

    def tendencies_nonlin(self, state_spect=None, old=None):
        """Compute the "nonlinear" tendencies."""
        oper = self.oper

        if state_spect is None:
            state_spect = self.state.state_spect

        if old is None:
            tendencies_fft = SetOfVariables(
                like=self.state.state_spect, info="tendencies_nonlin"
            )
        else:
            tendencies_fft = old

        w_fft = state_spect.get_var("w_fft")
        z_fft = state_spect.get_var("z_fft")

        F_fft = tendencies_fft.get_var("w_fft")
        tendencies_fft.set_var("z_fft", w_fft)

        mamp_zz = oper.monge_ampere_from_fft(z_fft, z_fft)
        chi_fft = -oper.invlaplacian_fft(oper.fft2(mamp_zz), order=4)
        mamp_zchi = oper.monge_ampere_from_fft(z_fft, chi_fft)
        Nw_fft = oper.fft2(mamp_zchi)
        lap2z_fft = oper.laplacian_fft(z_fft, order=4)
        F_fft[:] = -lap2z_fft + Nw_fft

        oper.dealiasing(tendencies_fft)

        # ratio = self.test_tendencies_nonlin(
        #     tendencies_fft, w_fft, z_fft, chi_fft)
        # print('ratio:', ratio)

        if self.params.forcing.enable:
            tendencies_fft += self.forcing.get_forcing()

        return tendencies_fft

    def compute_freq_diss(self):
        """Compute the dissipation frequencies with dissipation only for w."""
        f_d_w, f_d_hypo_w = super().compute_freq_diss()
        f_d = np.zeros_like(self.state.state_spect, dtype=np.float64)
        f_d_hypo = np.zeros_like(self.state.state_spect, dtype=np.float64)
        f_d[0] = f_d_w
        f_d_hypo[0] = f_d_hypo_w
        return f_d, f_d_hypo

    def test_tendencies_nonlin(
        self, tendencies_fft=None, w_fft=None, z_fft=None, chi_fft=None
    ):
        r"""Test if the tendencies conserves the total energy.

        We consider the conservative Föppl-von Kármán equations
        (without dissipation and forcing) written as

        .. math::

           \p_t z = F_z

           \p_t w = F_w

        We have:

        .. math::

           \p_t E_K(\mathbf{k}) = \mathcal{R} ( \hat F_w \hat w ^* )

           \p_t E_L(\mathbf{k}) = k^4 \mathcal{R} ( \hat F_z \hat z ^* )

           \p_t E_{NQ}(\mathbf{k}) =
           - \mathcal{R} ( \widehat{\{ F_z, z\}} \hat \chi ^* )

        Since the total energy is conserved, we should have

        .. math::

           \sum_{\mathbf{k}} \p_t E_K(\mathbf{k}) + \p_t E_L(\mathbf{k})
           + \p_t E_{NQ}(\mathbf{k}) = 0

        This function computes this quantities.

        """

        if tendencies_fft is None:
            tendencies_fft = self.tendencies_nonlin()
            w_fft = self.state.state_spect.get_var("w_fft")
            z_fft = self.state.state_spect.get_var("z_fft")
            chi_fft = self.state.get_var("chi_fft")

        F_w_fft = tendencies_fft.get_var("w_fft")
        F_z_fft = tendencies_fft.get_var("z_fft")

        K4 = self.oper.K4

        dt_E_K = np.real(F_w_fft * w_fft.conj())
        dt_E_L = K4 * np.real(F_z_fft * z_fft.conj())

        tmp = self.oper.monge_ampere_from_fft(F_z_fft, z_fft)
        tmp_fft = self.oper.fft2(tmp)

        dt_E_NQ = -np.real(tmp_fft * chi_fft.conj())

        T = dt_E_K + dt_E_L + dt_E_NQ

        norm = self.oper.sum_wavenumbers(abs(T))

        if norm < 1e-15:
            print("Only zeros in total energy tendency.")
            # print('(K+L)\n', dt_E_K+dt_E_L)
            # print('NQ\n', dt_E_NQ)
            return 0

        else:
            T = T / norm
            # print('ratio array\n', T)
            # print('(K+L)\n', (dt_E_K+dt_E_L)/norm)
            # print('NQ\n', dt_E_NQ/norm)
            return self.oper.sum_wavenumbers(T)


if __name__ == "__main__":

    np.set_printoptions(precision=2)

    import fluiddyn as fld

    params = Simul.create_default_params()

    params.short_name_type_run = "test"

    nh = 32
    Lh = 1.0
    params.oper.nx = nh
    params.oper.ny = nh
    params.oper.Lx = Lh
    params.oper.Ly = Lh
    params.oper.coef_dealiasing = 2.0 / 3

    delta_x = Lh / nh
    params.nu_8 = 2e1 * params.forcing.forcing_rate ** (1.0 / 3) * delta_x ** 8

    kmax = np.sqrt(2) * np.pi / delta_x
    deltat = 2 * np.pi / kmax ** 2 / 2

    params.time_stepping.USE_CFL = False
    params.time_stepping.deltat0 = deltat
    params.time_stepping.USE_T_END = False
    params.time_stepping.t_end = 1.0
    params.time_stepping.it_end = 10

    params.init_fields.type = "noise"
    params.init_fields.noise.velo_max = 1e-6
    # params.init_fields.path_file = ''

    params.forcing.enable = True
    params.forcing.type = "tcrandom"
    params.forcing.forcing_rate = 1e12
    params.forcing.nkmax_forcing = 5
    params.forcing.nkmin_forcing = 2
    params.forcing.tcrandom.time_correlation = 100 * deltat

    params.output.periods_print.print_stdout = 0.05

    params.output.periods_save.phys_fields = 5.0
    params.output.periods_save.spectra = 0.05
    params.output.periods_save.spatial_means = 10 * deltat
    params.output.periods_save.correl_freq = 1

    params.output.ONLINE_PLOT_OK = False
    params.output.period_refresh_plots = 0.05

    params.output.periods_plot.phys_fields = 0.1

    params.output.phys_fields.field_to_plot = "w"

    params.output.spectra.HAS_TO_PLOT_SAVED = True

    params.output.spatial_means.HAS_TO_PLOT_SAVED = True

    params.output.correl_freq.HAS_TO_PLOT_SAVED = False
    nb_times_compute = 200
    params.output.correl_freq.nb_times_compute = nb_times_compute
    params.output.correl_freq.coef_decimate = 1
    params.output.correl_freq.iomegas1 = np.linspace(
        1, nb_times_compute / 2 - 1, 6
    ).astype(int)

    sim = Simul(params)

    # sim.output.phys_fields.plot()
    sim.time_stepping.start()
    #    sim.output.phys_fields.plot()

    fld.show()



import matplotlib.pylab as plt
import glob
import numpy as np

from solveq2d import solveq2d


from create_figs_articles import CreateFigArticles


SAVE_FIG = False

create_fig = CreateFigArticles(
    short_name_article="SW1l", SAVE_FIG=SAVE_FIG, FOR_BEAMER=False, fontsize=19
)

dir_base = create_fig.path_base_dir + "/Results_SW1lw"

c = 40

resol = 240 * 2 ** 5

str_resol = repr(resol)
str_to_find_path = (
    dir_base + "/Pure_standing_waves_" + str_resol + "*/SE2D*c=" + repr(c)
) + "_*"
print(str_to_find_path)

paths_dir = glob.glob(str_to_find_path)

print(paths_dir)


sim = solveq2d.create_sim_plot_from_dir(paths_dir[0])

tmin = sim.output.spatial_means.first_saved_time()
tstatio = tmin + 4.0


tmax = 1000


key_var = "uy"

(
    pdf_timemean,
    values_inc_timemean,
    nb_rx_to_plot,
) = sim.output.increments.load_pdf_from_file(
    tmin=tmin, tmax=tmax, key_var=key_var
)


deltax = sim.param.Lx / sim.param.nx
rxs = np.array(sim.output.increments.rxs, dtype=np.float64) * deltax


# if 7680
rmin = 8 * deltax
rmax = 40 * deltax

# rmin = 0.6
# rmax = 3


# if 4096
# rmin = 8e-2
# rmin = 8*deltax
# rmax = 5e-1
# rmax = 50*deltax

# if 2048
# rmin = 1.5e-1
# rmin = 5*deltax
# rmax = 20*deltax
# rmax = 3e-1


condr = np.logical_and(rxs > rmin, rxs < rmax)


def expo_from_order(order, PLOT=False, PLOT_PDF=False):
    order = float(order)
    M_order = np.empty(rxs.shape)
    for irx in xrange(rxs.size):
        deltainc = values_inc_timemean[irx, 1] - values_inc_timemean[irx, 0]
        M_order[irx] = deltainc * np.sum(
            pdf_timemean[irx] * abs(values_inc_timemean[irx]) ** order
        )

        # M_order[irx] = np.abs(deltainc*np.sum(
        #     pdf_timemean[irx]
        #     *values_inc_timemean[irx]**order
        #     ))

    pol = np.polyfit(np.log(rxs[condr]), np.log(M_order[condr]), 1)
    expo = pol[0]

    print(f"order = {order:.2f} ; expo = {expo:.2f}")
    M_lin = np.exp((pol[1] + np.log(rxs[condr]) * pol[0]))

    if PLOT:
        fig, ax1 = sim.output.figure_axe()
        title = (
            "struct. function, solver "
            + sim.output.name_solver
            + f", nh = {sim.param.nx:5d}"
            + ", c = {:.4g}, f = {:.4g}".format(
                np.sqrt(sim.param.c2), sim.param.f
            )
        )

        ax1.set_xlabel("$r$")
        ax1.set_ylabel(r"$\langle |\delta v|^{" + f"{order:.2f}" + "}\\rangle/r$")

        ax1.set_title(title)
        ax1.hold(True)
        ax1.set_xscale("log")
        ax1.set_yscale("log")

        norm = rxs ** (1)

        ax1.plot(rxs, M_order / norm, "x-", linewidth=1)

        ax1.plot(rxs[condr], M_order[condr] / norm[condr], "x-r", linewidth=1)

        ax1.plot(rxs[condr], M_lin / norm[condr], "y", linewidth=2)

        l_1 = ax1.plot(rxs, rxs / norm, "k", linewidth=1)
        l_K41 = ax1.plot(rxs, rxs ** (order / 3) / norm, "k--", linewidth=1)

        cond = rxs < 4 * deltax
        temp = rxs ** (order) / norm
        l_smooth = ax1.plot(
            rxs[cond],
            temp[cond] / temp[0] * M_order[0] / norm[0],
            "k:",
            linewidth=1,
        )

        plt.text(1, 6 * 10 ** 0, f"order = {order:.2f}", fontsize=16)
        plt.text(0.11, 4 * 10 ** -1, f"expo = {expo:.3f}", fontsize=16)

        plt.rc("legend", numpoints=1)
        leg1 = ax1.legend(
            [l_smooth[0], l_K41[0], l_1[0]],
            ["smooth $r^q$", "K41 $r^{q/3}$", "shocks $r^1$"],
            loc=0
            # labelspacing = 0.2
        )

    if PLOT_PDF:
        fig, ax1 = sim.output.figure_axe()
        title = (
            "pdf increments, solver "
            + sim.output.name_solver
            + f", nh = {sim.param.nx:5d}"
            + f", c2 = {sim.param.c2:.4g}, f = {sim.param.f:.4g}"
        )

        ax1.set_title(title)
        ax1.hold(True)
        ax1.set_xscale("linear")
        ax1.set_yscale("linear")

        ax1.set_xlabel(key_var)
        ax1.set_ylabel(r"PDF x $\delta v^" + repr(order) + "$")

        colors = ["k", "y", "r", "b", "g", "m", "c"]

        irx_to_plot = [10, 50, 100]
        for irxp, irx in enumerate(irx_to_plot):

            val_inc = values_inc_timemean[irx]

            ax1.plot(
                val_inc,
                pdf_timemean[irx] * abs(val_inc) ** order,
                colors[irxp] + "x-",
                linewidth=1,
            )

    return expo


PLOT = False
PLOT = True
if PLOT:
    delta_order = 1
else:
    delta_order = 0.05


orders = np.arange(0.0, 6, delta_order)
expos = np.empty(orders.shape)

for iorder, order in enumerate(orders):
    expos[iorder] = expo_from_order(order, PLOT=PLOT, PLOT_PDF=False)

expos_K41 = orders / 3


fig, ax1 = sim.output.figure_axe()
title = (
    "intermittency, solver "
    + sim.output.name_solver
    + f", nh = {sim.param.nx:5d}"
    + ", c = {:.4g}, f = {:.4g}".format(np.sqrt(sim.param.c2), sim.param.f)
)

ax1.set_title(title)
ax1.hold(True)
ax1.set_xscale("linear")
ax1.set_yscale("linear")

ax1.set_xlabel("order q")
ax1.set_ylabel(r"$\zeta_q$")

ax1.plot(orders, expos)

ax1.plot(orders, expos_K41, "k--")

solveq2d.show()



"""Base simulations (:mod:`fluidsim.base.solvers`)
========================================================


Provides:

.. autosummary::
   :toctree:

   base
   pseudo_spect
   finite_diff
   info_base

"""



"""Base solver (:mod:`fluidsim.base.solvers.base`)
==================================================

Provides:

.. autoclass:: InfoSolverBase
   :members:
   :private-members:

.. autoclass:: SimulBase
   :members:
   :private-members:

"""
from time import time

import numpy as np
from fluidsim_core.solver import SimulCore

from ..setofvariables import SetOfVariables
from .info_base import InfoSolverBase


class SimulBase(SimulCore):
    """Represent a solver.

    This is the main base class which is inherited by the other
    simulation classes.

    A :class:`SimulBase` object contains at least one object of the
    classes:

    - :class:`fluidsim.base.params.Parameters`
    - :class:`fluidsim.base.time_stepping.TimeSteppingBase`
    - :class:`fluidsim.operators.operators.Operators`
    - :class:`fluidsim.base.state.StateBase`

    Parameters
    ----------

    params : :class:`fluidsim.base.params.Parameters`
        Parameters for the simulation.

    info_solver : :class:`fluidsim.base.solvers.info_base.InfoSolverBase`
        Information about the particular solver.

    """

    InfoSolver = InfoSolverBase

    @staticmethod
    def _complete_params_with_default(params):
        """A static method used to complete the *params* container."""
        attribs = {
            "short_name_type_run": "",
            "NEW_DIR_RESULTS": True,
            "ONLY_COARSE_OPER": False,
            # Physical parameters:
            "nu_2": 0.0,
        }
        params._set_attribs(attribs)
        params._set_doc(
            """
short_name_type_run: str

    A short name of the simulation used to create the directory name.

NEW_DIR_RESULTS: bool

    To be used only when loading a simulation. If True (default), a new directory
    is created to contain the results of the simulation. If False, the results of
    the simulation are appended in the old directory.

ONLY_COARSE_OPER: bool

    To be used only when loading a simulation. If True (not default), the operator
    is created with a very small resolution. It is very fast but then it can not
    be used to process data.

nu_2: float (default = 0.)

    Viscosity coefficient. Used in particular in the method
    :func:`fluidsim.base.solvers.pseudo_spect.SimulBasePseudoSpectral.compute_freq_diss`).

"""
        )

    @classmethod
    def create_default_params(cls):
        return super().create_default_params()

    def __enter__(self):
        if not hasattr(self, "_end_of_simul") or self._end_of_simul:
            self.time_stepping._time_beginning_simul = time()
            self._end_of_simul = False

        return self

    def __exit__(self, *args):
        if not self._end_of_simul:
            total_time_simul = time() - self.time_stepping._time_beginning_simul
            self.time_stepping.time_simul_in_sec = total_time_simul
            self.output.end_of_simul(total_time_simul)
            self._end_of_simul = True

    def __init__(self, params):
        super().__init__(params)

        # np.seterr(invalid='raise')
        # np.seterr(over='raise')
        np.seterr(all="warn")
        np.seterr(under="ignore")

        dict_classes = self.info_solver.import_classes()

        # initialization operators and grid
        Operators = dict_classes["Operators"]
        self.oper = Operators(params=params)

        # initialization output
        Output = dict_classes["Output"]
        self.output = Output(self)

        self.output.print_stdout(
            "*************************************\nProgram fluidsim"
        )

        # output.print_memory_usage(
        #     'Memory usage after creating operator (equiv. seq.)')

        # initialisation object variables
        State = dict_classes["State"]
        self.state = State(self)

        # initialisation time stepping
        TimeStepping = dict_classes["TimeStepping"]
        self.time_stepping = TimeStepping(self)

        # initialisation fields (and time if needed)
        InitFields = dict_classes["InitFields"]
        self.init_fields = InitFields(self)
        self.init_fields()

        # initialisation forcing
        self.is_forcing_enabled = False
        try:
            params.forcing
        except AttributeError:
            pass
        else:
            if params.forcing.enable:
                self.is_forcing_enabled = True
                Forcing = dict_classes["Forcing"]
                self.forcing = Forcing(self)
        # we can not yet compute a forcing...
        # self.forcing.compute()

        # complete the initialisation of the object output
        self.output.post_init()

        # if enabled, preprocesses flow parameters such as viscosity and
        # forcing based on initialized fields
        if "Preprocess" in dict_classes:
            Preprocess = dict_classes["Preprocess"]
            self.preprocess = Preprocess(self)
            self.preprocess()

    def tendencies_nonlin(self, variables=None, old=None):
        """Return a null SetOfVariables object."""
        if old is None:
            tendencies = SetOfVariables(
                like=self.state.state_phys, info="tendencies_nonlin"
            )
        else:
            tendencies = old
        tendencies.initialize(value=0.0)
        return tendencies


Simul = SimulBase


if __name__ == "__main__":

    params = Simul.create_default_params()

    params.short_name_type_run = "test"
    params.time_stepping.USE_CFL = False
    params.time_stepping.t_end = 2.0
    params.time_stepping.deltat0 = 0.1

    sim = Simul(params)
    sim.time_stepping.start()



"""Spherical harmonics operators (:mod:`fluidsim.operators.sphericalharmo`)
===========================================================================


"""
import numpy as np

from pyshtools.constants import Earth

from fluiddyn.util.compat import cached_property

from fluidsht.sht2d.operators import OperatorsSphereHarmo2D as _Operator

# from fluiddyn.calcul.sphericalharmo import EasySHT as _Operator

from fluidsim.base.params import Parameters


class OperatorsSphericalHarmonics(_Operator):
    @staticmethod
    def _complete_params_with_default(params):
        f"""This static method is used to complete the *params* container.
        Default parameters

        - omega (solid body rotation speed)
        - radius (of the sphere)

        are initialized as per *{Earth.wgs84.r3.reference}*

        """

        attribs = {
            "lmax": 15,
            "nlat": None,
            "nlon": None,
            "omega": Earth.wgs84.omega.value,
            "radius": Earth.wgs84.r3.value,
        }
        params._set_child("oper", attribs=attribs)

    def __init__(self, params=None):
        self.axes = ("lat", "lon")
        if params is None:
            params = Parameters(tag="params")
            self.__class__._complete_params_with_default(params)

        self.params = params = params.oper
        super().__init__(
            lmax=params.lmax,
            nlat=params.nlat,
            nlon=params.nlon,
            radius=params.radius,
        )

    @cached_property
    def f_radial(self):
        r"""The vertical / radial component of the solid body rotation vector
        pointed along axis of rotation. Useful for calculation of coriolis
        term where :math:`f = 2 \Omega \sin \theta`.

        """
        return 2.0 * self.params.omega * np.sin(self.LATS)

    def get_grid1d_seq(self, axe="lat"):
        if axe not in self.axes:
            raise ValueError

        # TODO: implement
        # if self.params.ONLY_COARSE_OPER:
        # set_grid based on params.nlat and params.nlon and other flags
        # rebuild lats and lons
        # else:
        return getattr(self, axe + "s")


if __name__ == "__main__":

    oper = OperatorsSphericalHarmonics()



from fluidsim.solvers.ns2d.output.print_stdout import PrintStdOutNS2D


class PrintStdOutAD1D(PrintStdOutNS2D):
    """Used to print in both the stdout and the stdout.txt file, and also
    to print simple info on the current state of the simulation.

    """
